#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

from qiling.const import *
from qiling.os.const import *
from .const import *
from .utils import *
from .type64 import *
from .fncc import *

@dxeapi(params={
    "a0": POINTER, #POINTER_T(struct_EFI_TIME)
    "a1": POINTER, #POINTER_T(struct_EFI_TIME_CAPABILITIES)
})
def hook_GetTime(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(struct_EFI_TIME)
})
def hook_SetTime(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(ctypes.c_ubyte)
    "a1": POINTER, #POINTER_T(ctypes.c_ubyte)
    "a2": POINTER, #POINTER_T(struct_EFI_TIME)
})
def hook_GetWakeupTime(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(struct_EFI_TIME)
})
def hook_SetWakeupTime(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": UINT,
    "a3": POINTER, #POINTER_T(struct_EFI_MEMORY_DESCRIPTOR)
})
def hook_SetVirtualAddressMap(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_ConvertPointer(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "VariableName": WSTRING,
    "VendorGuid": GUID,
    "Attributes": POINTER,
    "DataSize": POINTER,
    "Data": POINTER
})
def hook_GetVariable(ql, address, params):
    if params['VariableName'] in ql.env:
        var = ql.env[params['VariableName']]
        read_len = read_int64(ql, params['DataSize'])
        if params['Attributes'] != 0:
            write_int64(ql, params['Attributes'], 0)
        write_int64(ql, params['DataSize'], len(var))
        if read_len < len(var):
            return EFI_BUFFER_TOO_SMALL
        if params['Data'] != 0:
            ql.mem.write(params['Data'], var)
        return EFI_SUCCESS
    return EFI_NOT_FOUND

@dxeapi(params={
    "VariableNameSize": POINTER, #POINTER_T(ctypes.c_uint64)
    "VariableName": POINTER, #POINTER_T(ctypes.c_uint16)
    "VendorGuid": GUID,
})
def hook_GetNextVariableName(ql, address, params):
    name_size = read_int64(ql, params["VariableNameSize"])
    last_name = ql.os.read_wstring(params["VariableName"])
    vars = ql.env['Names'] # This is a list of variable names in correct order.
    if last_name in vars and vars.index(last_name) < len(vars) - 1:
        new_name = vars[vars.index(last_name)+1]
        if (len(new_name)+1)*2 > name_size:
            return EFI_BUFFER_TOO_SMALL
        vn_ptr = params["VariableName"]
        for char in new_name:
            ql.mem.write(vn_ptr, char)
            vn_ptr += 1
            ql.mem.write(vn_ptr, '\x00')
            vn_ptr += 1
        ql.mem.write(vn_ptr, '\x00\x00')

    return EFI_INVALID_PARAMETER

@dxeapi(params={
    "VariableName": WSTRING, #POINTER_T(ctypes.c_uint16)
    "VendorGuid": GUID,
    "Attributes": UINT,
    "DataSize": ULONGLONG,
    "Data": POINTER, #POINTER_T(None)
})
def hook_SetVariable(ql, address, params):
    ql.env[params['VariableName']] = bytes(ql.mem.read(params['Data'], params['DataSize']))
    return EFI_SUCCESS

@dxeapi(params={
    "Count": POINTER, #POINTER_T(ctypes.c_uint32)
})
def hook_GetNextHighMonotonicCount(ql, address, params):
    ql.os.monotonic_count += 0x0000000100000000
    hmc = ql.os.monotonic_count
    hmc = (hmc >> 32) & 0xffffffff
    write_int32(ql, params["Count"], hmc)
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": POINTER, #POINTER_T(None)
})
def hook_ResetSystem(ql, address, params):
    ql.nprint(f'hook_ResetSystem')
    ql.emu_stop()
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(POINTER_T(struct_EFI_CAPSULE_HEADER))
    "a1": ULONGLONG,
    "a2": ULONGLONG,
})
def hook_UpdateCapsule(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(POINTER_T(struct_EFI_CAPSULE_HEADER))
    "a1": ULONGLONG,
    "a2": POINTER, #POINTER_T(ctypes.c_uint64)
    "a3": POINTER, #POINTER_T(enum_73)
})
def hook_QueryCapsuleCapabilities(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": UINT,
    "a1": POINTER, #POINTER_T(ctypes.c_uint64)
    "a2": POINTER, #POINTER_T(ctypes.c_uint64)
    "a3": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_QueryVariableInfo(ql, address, params):
    return EFI_SUCCESS

def hook_EFI_RUNTIME_SERVICES(ql, start_ptr):
    efi_runtime_services = EFI_RUNTIME_SERVICES()
    ptr = start_ptr
    pointer_size = 8
    efi_runtime_services.GetTime = ptr
    ql.hook_address(hook_GetTime, ptr)
    ptr += pointer_size
    efi_runtime_services.SetTime = ptr
    ql.hook_address(hook_SetTime, ptr)
    ptr += pointer_size
    efi_runtime_services.GetWakeupTime = ptr
    ql.hook_address(hook_GetWakeupTime, ptr)
    ptr += pointer_size
    efi_runtime_services.SetWakeupTime = ptr
    ql.hook_address(hook_SetWakeupTime, ptr)
    ptr += pointer_size
    efi_runtime_services.SetVirtualAddressMap = ptr
    ql.hook_address(hook_SetVirtualAddressMap, ptr)
    ptr += pointer_size
    efi_runtime_services.ConvertPointer = ptr
    ql.hook_address(hook_ConvertPointer, ptr)
    ptr += pointer_size
    efi_runtime_services.GetVariable = ptr
    ql.hook_address(hook_GetVariable, ptr)
    ptr += pointer_size
    efi_runtime_services.GetNextVariableName = ptr
    ql.hook_address(hook_GetNextVariableName, ptr)
    ptr += pointer_size
    efi_runtime_services.SetVariable = ptr
    ql.hook_address(hook_SetVariable, ptr)
    ptr += pointer_size
    efi_runtime_services.GetNextHighMonotonicCount = ptr
    ql.hook_address(hook_GetNextHighMonotonicCount, ptr)
    ptr += pointer_size
    efi_runtime_services.ResetSystem = ptr
    ql.hook_address(hook_ResetSystem, ptr)
    ptr += pointer_size
    efi_runtime_services.UpdateCapsule = ptr
    ql.hook_address(hook_UpdateCapsule, ptr)
    ptr += pointer_size
    efi_runtime_services.QueryCapsuleCapabilities = ptr
    ql.hook_address(hook_QueryCapsuleCapabilities, ptr)
    ptr += pointer_size
    efi_runtime_services.QueryVariableInfo = ptr
    ql.hook_address(hook_QueryVariableInfo, ptr)
    ptr += pointer_size
    return (ptr, efi_runtime_services)




#!/usr/bin/env python3

import cmd
from functools import partial

from qiling import *
from qiling.const import *
from qiling.debugger import QlDebugger

from .frontend import context_printer, context_reg, context_asm, examine_mem, color
from .utils import parse_int, handle_bnj, is_thumb, diff_snapshot_save, diff_snapshot_restore, CODE_END


class QlQdb(cmd.Cmd, QlDebugger):
    def __init__(self, ql, rr=False):

        self._ql = ql
        self.prompt = "(Qdb) "
        self.breakpoints = {}
        self._saved_states = None
        if rr:
            self._states_list = [None]

        super().__init__()

        # setup a breakpoint at entry point
        self._ql.hook_address(self._attach, self._ql.loader.entry_point)

    def _attach(self, ql, *args, **kwargs):
        print(color.RED, "[+] Qdb attached", color.END, sep="")
        print(color.RED, "[!] All hooks of qiling instance will be disabled in Qdb", color.END, sep="")

        # clear all hooks
        for i in ql._addr_hook_fuc.keys():
            ql.uc.hook_del(ql._addr_hook_fuc[i])

        self.interactive()

    def interactive(self, *args):
        self.cmdloop()

    def emptyline(self, *args):
        """
        repeat last command
        """
        _lastcmd = getattr(self, "do_" + self.lastcmd, None)
        if _lastcmd:
            return _lastcmd()

    def del_breakpoint(self, address):
        """
        handle internal breakpoint removing operation
        """
        _bp = self.breakpoints.pop(address, None)
        if _bp:
            _bp["hook"].remove()

    def set_breakpoint(self, address, _is_temp=False):
        """
        handle internal breakpoint adding operation
        """
        _bp_func = partial(self._breakpoint_handler, _is_temp=_is_temp)

        _hook = self._ql.hook_address(_bp_func, address)
        self.breakpoints.update({address: {"hook": _hook, "hitted": False, "temp": _is_temp}})

        if _is_temp == False:
            print("Breakpoint at 0x%08x" % address)

    def _breakpoint_handler(self, ql, _is_temp=False):
        """
        handle all breakpoints
        """
        _cur_addr = ql.reg.arch_pc

        if _is_temp: # remove temporary breakpoint
            self.del_breakpoint(_cur_addr)
        else:
            if self.breakpoints.get(_cur_addr)["hitted"]:
                return

            print("hit breakpoint at 0x%08x" % _cur_addr)
            self.breakpoints.get(_cur_addr)["hitted"] = True

        self.do_context()
        self._ql.emu_stop()

    def do_context(self, *args):
        """
        show context information for current location
        """
        context_reg(self._ql, self._saved_states)
        context_asm(self._ql, self._ql.reg.arch_pc, 4)

    def do_run(self, *args):
        """
        launch qiling instance
        """

        entry = self._ql.loader.entry_point

        self.run(entry)

    def run(self, address=None):
        """
        handle qiling instance launching
        """

        if address is None:
            return

        # for arm thumb mode
        if self._ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB) and is_thumb(self._ql.reg.cpsr):
            address |= 1

        self._ql.emu_start(address, 0)

    def do_backward(self, *args):

        if getattr(self, "_states_list", None) is None or self._states_list[-1] is None:
            print("there is no way back !!!")
        else:
            print("step backward ~")
            current_state_dicts = self._ql.save(cpu_context=True, mem=True, reg=False, fd=False)
            self._ql.restore(diff_snapshot_restore(current_state_dicts, self._states_list.pop()))
            self.do_context()

    def do_step(self, *args):
        """
        execute one instruction at a time
        """

        if self._ql is None:
            print("The program is not being run.")

        else:
            self._saved_states = dict(filter(lambda d: isinstance(d[0], str), self._ql.reg.save().items()))

            if getattr(self, "_states_list", None) is not None:
                current_state_dicts = self._ql.save(cpu_context=True, mem=True, reg=False, fd=False)
                self._states_list.append(diff_snapshot_save(current_state_dicts, self._states_list[-1]))

            _cur_addr = self._ql.reg.arch_pc

            next_stop = handle_bnj(self._ql, _cur_addr)

            if next_stop is CODE_END:
                return True

            # whether bp placed already
            if self.breakpoints.get(next_stop, None):
                self.breakpoints.get(next_stop)["hitted"] = False

            else:
                self.set_breakpoint(next_stop, _is_temp=True)

            self.run(_cur_addr)

    def do_start(self, *args):
        """
        pause at entry point by setting a temporary breakpoint on it
        """
        entry = self._ql.loader.entry_point  # ld.so
        # entry = self._ql.loader.elf_entry # .text of binary

        if self._ql.archtype in (QL_ARCH.ARM, QL_ARCH.ARM_THUMB) and entry & 1:
            entry -= 1

        self.set_breakpoint(entry, _is_temp=True)
        self.do_run()

    def do_breakpoint(self, address):
        """
        set breakpoint on specific address
        """
        if address:
            baddr = parse_int(address)
            self.set_breakpoint(baddr)

    def do_continue(self, *args):
        """
        continue execution till next breakpoint or the end
        """
        if self._ql is not None:
            _cur_addr = self._ql.reg.arch_pc
            print("continued from 0x%08x" % _cur_addr)

            self.run(_cur_addr)

    def do_examine(self, args):
        """
        read data from memory of qiling instance
        """

        _args = args.split()

        if len(_args) == 1:
            _xaddr = parse_int(_args[0])
            _count = 1

        elif len(_args) == 2:
            _xaddr, _count = _args
            _xaddr = parse_int(_xaddr)
            _count = parse_int(_count)

        else:
            print("wrong format\nUsage: x ADDRESS [SIZE]")
            return

        examine_mem(self._ql, _xaddr, _count)

    def do_context(self, *args):
        """
        show context information for current location
        """
        context_reg(self._ql, self._saved_states)
        context_asm(self._ql, self._ql.reg.arch_pc, 4)

    def do_show(self, *args):
        """
        show some runtime informations
        """
        self._ql.mem.show_mapinfo()
        print("Qdb:", [(hex(idx), val) for idx, val in self.breakpoints.items()])
        print("internal:", [(hex(idx), val) for idx, val in self._ql._addr_hook.items()])

    def do_disassemble(self, address):
        """
        disassemble instructions from address specified
        """
        context_asm(self._ql, parse_int(address), 4)

    def do_shell(self, *command):
        """
        run python code,also a space between exclamation mark and command was necessary
        """
        try:
            print(eval(*command))
        except:
            print("something went wrong")

    def do_quit(self, *args):
        """
        exit Qdb
        """
        return True

    do_r = do_run
    do_s = do_step
    do_q = do_quit
    do_x = do_examine
    do_c = do_continue
    do_b = do_breakpoint
    do_p = do_backward
    do_dis = do_disassemble
    


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 
import struct
from unicorn.x86_const import *
from qiling.os.utils import *
from qiling.os.macos.utils import *

DWORD = 1
UINT = 1
ULONG = 1
INT = 1
BOOL = 1
SIZE_T = 1
BYTE = 1
CHAR = 1
UCHAR = 1
USHORT = 1
SHORT = 1
ULONGLONG = 2
POINTER = 3
STRING = 4
BOOLEAN = 8
BOOL = 8


def _get_param_by_index(ql, index):
    return get_params_by_index(ql, index)


def get_params_by_index(ql, index):
    reg_list = [UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_RCX, UC_X86_REG_R8, UC_X86_REG_R9]
    if index < 6:
        return ql.uc.reg_read(reg_list[index])

    index -= 6
    # skip ret_addr
    return ql.stack_read((index + 1) * 8)


def _x8664_get_args(ql, number):
    reg_list = [UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_RCX, UC_X86_REG_R8, UC_X86_REG_R9]
    arg_list = []
    reg_num = number
    if reg_num > 6:
        reg_num = 6

    number -= reg_num
    # get args in registers first
    for i in reg_list[:reg_num]:
        arg_list.append(ql.uc.reg_read(i))

    # the rest args are from stack
    for i in range(number):
        # skip ret_addr and 32 byte home space
        arg_list.append(ql.stack_read((i + 7) * 8))

    if reg_num == 1:
        return arg_list[0]
    else:
        return arg_list


def set_function_params(ql, in_params, out_params, index=0):
    for each in in_params:
        if in_params[each] in (DWORD, POINTER, ULONGLONG):
            out_params[each] = _get_param_by_index(ql, index)
        elif in_params[each] == STRING:
            ptr = _get_param_by_index(ql, index)
            if ptr == 0:
                out_params[each] = 0
            else:
                out_params[each] = macho_read_string(ql, ptr, 0x1000)
        elif in_params[each] in (BOOLEAN, BOOL):
            ptr = _get_param_by_index(ql, index)
            if ptr == 1:
                out_params[each] = "True"
            else:
                out_params[each] = "False"

        index += 1

    return index


def get_function_param(ql, number):
    return _x8664_get_args(ql, number)


def set_return_value(ql, ret):
    ql.uc.reg_write(UC_X86_REG_RAX, ret)


def get_return_value(ql):
    return ql.uc.reg_read(UC_X86_REG_RAX)


def __x86_cc(ql, passthru, param_num, params, func, args, kwargs):
    # read params
    if params is not None:
        param_num = set_function_params(ql, params, args[2])

    # call function
    result = func(*args, **kwargs)

    # set return value
    if result is not None:
        set_return_value(ql, result)

    # print
    print_function(ql, passthru, args[1], func.__name__, args[2], result)
    return result, param_num


def x8664_fastcall(ql, passthru, param_num, params, func, args, kwargs):
    result, param_num = __x86_cc(ql, passthru, param_num, params, func, args, kwargs)

    if not passthru and ql.os.RUN is True:
        try:
            ql.reg.arch_pc = ql.stack_pop()
        except UcError:
            raise

    return result


def macos_kernel_api(param_num=None, params=None, passthru=False):
    def decorator(func):
        def wrapper(*args, **kwargs):
            ql = args[0]
            return x8664_fastcall(ql, passthru, param_num, params, func, args, kwargs)
        return wrapper
    return decorator




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

import struct
from qiling.const import *

def dxeapi(param_num=None, params=None):
    def decorator(func):
        def wrapper(*args, **kwargs):
            ql = args[0]
            arg = (ql, ql.reg.arch_pc, {})
            f = func

            if func.__name__ in ql.loader.user_defined_api:
                f = ql.loader.user_defined_api[func.__name__]
            
            if func.__name__ in ql.loader.user_defined_api_onenter:
                ql.os.winapi_func_onenter = ql.loader.user_defined_api_onenter[func.__name__]
            else:    
                ql.os.winapi_func_onenter = None

            if func.__name__ in ql.loader.user_defined_api_onexit:
                ql.os.winapi_func_onexit = ql.loader.user_defined_api_onexit[func.__name__]
            else:    
                ql.os.winapi_func_onexit = None                

            return ql.os.x8664_fastcall(param_num, params, f, arg, kwargs)

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
#
# TARGET arch is: []
# WORD_SIZE is: 8
# POINTER_SIZE is: 8
# LONGDOUBLE_SIZE is: 16
#
import ctypes


c_int128 = ctypes.c_ubyte*16
c_uint128 = c_int128
void = None
if ctypes.sizeof(ctypes.c_longdouble) == 16:
    c_long_double_t = ctypes.c_longdouble
else:
    c_long_double_t = ctypes.c_ubyte*16

# if local wordsize is same as target, keep ctypes pointer function.
# required to access _ctypes
import _ctypes
# Emulate a pointer class using the approriate c_int32/c_int64 type
# The new class should have :
# ['__module__', 'from_param', '_type_', '__dict__', '__weakref__', '__doc__']
# but the class should be submitted to a unique instance for each base type
# to that if A == B, POINTER_T(A) == POINTER_T(B)
ctypes._pointer_t_type_cache = {}
def POINTER_T(pointee):
    # a pointer should have the same length as LONG
    fake_ptr_base_type = ctypes.c_uint64 
    # specific case for c_void_p
    if pointee is None: # VOID pointer type. c_void_p.
        pointee = type(None) # ctypes.c_void_p # ctypes.c_ulong
        clsname = 'c_void'
    else:
        clsname = pointee.__name__
    if clsname in ctypes._pointer_t_type_cache:
        return ctypes._pointer_t_type_cache[clsname]
    # make template
    _class = type('LP_%d_%s'%(8, clsname), (fake_ptr_base_type,),{}) 
    ctypes._pointer_t_type_cache[clsname] = _class
    return _class



undefined = ctypes.c_ubyte
ImageBaseOffset32 = ctypes.c_uint32
byte = ctypes.c_ubyte
dword = ctypes.c_uint32
longlong = ctypes.c_int64
qword = ctypes.c_uint64
uchar = ctypes.c_ubyte
uint = ctypes.c_uint32
ulonglong = ctypes.c_uint64
undefined1 = ctypes.c_ubyte
undefined2 = ctypes.c_uint16
undefined4 = ctypes.c_uint32
undefined8 = ctypes.c_uint64
ushort = ctypes.c_uint16
word = ctypes.c_uint16
class struct__EFI_MM_SYSTEM_TABLE(ctypes.Structure):
    pass

class struct_GUID(ctypes.Structure):
    pass

class struct__EFI_MM_CPU_IO_PROTOCOL(ctypes.Structure):
    pass

class struct_EFI_MM_IO_ACCESS(ctypes.Structure):
    pass


# values for enumeration 'enum_859'
enum_859__enumvalues = {
    1: 'MM_IO_UINT16',
    2: 'MM_IO_UINT32',
    3: 'MM_IO_UINT64',
    0: 'MM_IO_UINT8',
}
MM_IO_UINT16 = 1
MM_IO_UINT32 = 2
MM_IO_UINT64 = 3
MM_IO_UINT8 = 0
enum_859 = ctypes.c_int # enum
struct_EFI_MM_IO_ACCESS._pack_ = True # source:False
struct_EFI_MM_IO_ACCESS._functions_ = []
struct_EFI_MM_IO_ACCESS._fields_ = [
    ('Read', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL), enum_859, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(None)))),
    ('Write', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL), enum_859, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(None)))),
]

struct_EFI_MM_IO_ACCESS._functions_.append(("Read",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL)', 'enum_859', 'ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(None)']))
struct_EFI_MM_IO_ACCESS._functions_.append(("Write",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL)', 'enum_859', 'ctypes.c_uint64', 'ctypes.c_uint64', 'POINTER_T(None)']))
struct__EFI_MM_CPU_IO_PROTOCOL._pack_ = True # source:False
struct__EFI_MM_CPU_IO_PROTOCOL._functions_ = []
struct__EFI_MM_CPU_IO_PROTOCOL._fields_ = [
    ('Mem', struct_EFI_MM_IO_ACCESS),
    ('Io', struct_EFI_MM_IO_ACCESS),
]

EFI_MM_CPU_IO_PROTOCOL = struct__EFI_MM_CPU_IO_PROTOCOL
class struct_EFI_CONFIGURATION_TABLE(ctypes.Structure):
    pass

struct_GUID._pack_ = True # source:False
struct_GUID._functions_ = []
struct_GUID._fields_ = [
    ('Data1', ctypes.c_uint32),
    ('Data2', ctypes.c_uint16),
    ('Data3', ctypes.c_uint16),
    ('Data4', ctypes.c_ubyte * 8),
]

EFI_GUID = struct_GUID
struct_EFI_CONFIGURATION_TABLE._pack_ = True # source:False
struct_EFI_CONFIGURATION_TABLE._functions_ = []
struct_EFI_CONFIGURATION_TABLE._fields_ = [
    ('VendorGuid', EFI_GUID),
    ('VendorTable', POINTER_T(None)),
]


# values for enumeration 'enum_494'
enum_494__enumvalues = {
    2: 'AllocateAddress',
    0: 'AllocateAnyPages',
    1: 'AllocateMaxAddress',
    3: 'MaxAllocateType',
}
AllocateAddress = 2
AllocateAnyPages = 0
AllocateMaxAddress = 1
MaxAllocateType = 3
enum_494 = ctypes.c_int # enum

# values for enumeration 'enum_16'
enum_16__enumvalues = {
    10: 'EfiACPIMemoryNVS',
    9: 'EfiACPIReclaimMemory',
    3: 'EfiBootServicesCode',
    4: 'EfiBootServicesData',
    7: 'EfiConventionalMemory',
    1: 'EfiLoaderCode',
    2: 'EfiLoaderData',
    15: 'EfiMaxMemoryType',
    11: 'EfiMemoryMappedIO',
    12: 'EfiMemoryMappedIOPortSpace',
    13: 'EfiPalCode',
    14: 'EfiPersistentMemory',
    0: 'EfiReservedMemoryType',
    5: 'EfiRuntimeServicesCode',
    6: 'EfiRuntimeServicesData',
    8: 'EfiUnusableMemory',
}
EfiACPIMemoryNVS = 10
EfiACPIReclaimMemory = 9
EfiBootServicesCode = 3
EfiBootServicesData = 4
EfiConventionalMemory = 7
EfiLoaderCode = 1
EfiLoaderData = 2
EfiMaxMemoryType = 15
EfiMemoryMappedIO = 11
EfiMemoryMappedIOPortSpace = 12
EfiPalCode = 13
EfiPersistentMemory = 14
EfiReservedMemoryType = 0
EfiRuntimeServicesCode = 5
EfiRuntimeServicesData = 6
EfiUnusableMemory = 8
enum_16 = ctypes.c_int # enum

# values for enumeration 'enum_498'
enum_498__enumvalues = {
    0: 'EFI_NATIVE_INTERFACE',
}
EFI_NATIVE_INTERFACE = 0
enum_498 = ctypes.c_int # enum

# values for enumeration 'enum_500'
enum_500__enumvalues = {
    0: 'AllHandles',
    2: 'ByProtocol',
    1: 'ByRegisterNotify',
}
AllHandles = 0
ByProtocol = 2
ByRegisterNotify = 1
enum_500 = ctypes.c_int # enum
class struct_EFI_TABLE_HEADER(ctypes.Structure):
    _pack_ = True # source:False
    _functions_ = []
    _fields_ = [
    ('Signature', ctypes.c_uint64),
    ('Revision', ctypes.c_uint32),
    ('HeaderSize', ctypes.c_uint32),
    ('CRC32', ctypes.c_uint32),
    ('Reserved', ctypes.c_uint32),
     ]

struct__EFI_MM_SYSTEM_TABLE._pack_ = True # source:False
struct__EFI_MM_SYSTEM_TABLE._functions_ = []
struct__EFI_MM_SYSTEM_TABLE._fields_ = [
    ('Hdr', struct_EFI_TABLE_HEADER),
    ('MmFirmwareVendor', POINTER_T(ctypes.c_uint16)),
    ('MmFirmwareRevision', ctypes.c_uint32),
    ('PADDING_0', ctypes.c_ubyte * 4),
    ('MmInstallConfigurationTable', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_SYSTEM_TABLE), POINTER_T(struct_GUID), POINTER_T(None), ctypes.c_uint64))),
    ('MmIo', EFI_MM_CPU_IO_PROTOCOL),
    ('MmAllocatePool', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_16, ctypes.c_uint64, POINTER_T(POINTER_T(None))))),
    ('MmFreePool', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None)))),
    ('MmAllocatePages', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_494, enum_16, ctypes.c_uint64, POINTER_T(ctypes.c_uint64)))),
    ('MmFreePages', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))),
    ('MmStartupThisAp', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.CFUNCTYPE(None, POINTER_T(None))), ctypes.c_uint64, POINTER_T(None)))),
    ('CurrentlyExecutingCpu', ctypes.c_uint64),
    ('NumberOfCpus', ctypes.c_uint64),
    ('CpuSaveStateSize', POINTER_T(ctypes.c_uint64)),
    ('CpuSaveState', POINTER_T(POINTER_T(None))),
    ('NumberOfTableEntries', ctypes.c_uint64),
    ('MmConfigurationTable', POINTER_T(struct_EFI_CONFIGURATION_TABLE)),
    ('MmInstallProtocolInterface', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(POINTER_T(None)), POINTER_T(struct_GUID), enum_498, POINTER_T(None)))),
    ('MmUninstallProtocolInterface', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID), POINTER_T(None)))),
    ('MmHandleProtocol', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID), POINTER_T(POINTER_T(None))))),
    ('MmRegisterProtocolNotify', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None))), POINTER_T(POINTER_T(None))))),
    ('MmLocateHandle', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_500, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(None))))),
    ('MmLocateProtocol', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(POINTER_T(None))))),
    ('MmiManage', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))),
    ('MmiHandlerRegister', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64))), POINTER_T(struct_GUID), POINTER_T(POINTER_T(None))))),
    ('MmiHandlerUnRegister', POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None)))),
]

struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmInstallConfigurationTable",['ctypes.c_uint64', 'POINTER_T(struct__EFI_MM_SYSTEM_TABLE)', 'POINTER_T(struct_GUID)', 'POINTER_T(None)', 'ctypes.c_uint64']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmAllocatePool",['ctypes.c_uint64', 'enum_16', 'ctypes.c_uint64', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmFreePool",['ctypes.c_uint64', 'POINTER_T(None)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmAllocatePages",['ctypes.c_uint64', 'enum_494', 'enum_16', 'ctypes.c_uint64', 'POINTER_T(ctypes.c_uint64)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmFreePages",['ctypes.c_uint64', 'ctypes.c_uint64', 'ctypes.c_uint64']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmStartupThisAp",['ctypes.c_uint64', 'POINTER_T(ctypes.CFUNCTYPE(None, POINTER_T(None)))', 'ctypes.c_uint64', 'POINTER_T(None)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmInstallProtocolInterface",['ctypes.c_uint64', 'POINTER_T(POINTER_T(None))', 'POINTER_T(struct_GUID)', 'enum_498', 'POINTER_T(None)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmUninstallProtocolInterface",['ctypes.c_uint64', 'POINTER_T(None)', 'POINTER_T(struct_GUID)', 'POINTER_T(None)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmHandleProtocol",['ctypes.c_uint64', 'POINTER_T(None)', 'POINTER_T(struct_GUID)', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmRegisterProtocolNotify",['ctypes.c_uint64', 'POINTER_T(struct_GUID)', 'POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None)))', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmLocateHandle",['ctypes.c_uint64', 'enum_500', 'POINTER_T(struct_GUID)', 'POINTER_T(None)', 'POINTER_T(ctypes.c_uint64)', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmLocateProtocol",['ctypes.c_uint64', 'POINTER_T(struct_GUID)', 'POINTER_T(None)', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmiManage",['ctypes.c_uint64', 'POINTER_T(struct_GUID)', 'POINTER_T(None)', 'POINTER_T(None)', 'POINTER_T(ctypes.c_uint64)']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmiHandlerRegister",['ctypes.c_uint64', 'POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))', 'POINTER_T(struct_GUID)', 'POINTER_T(POINTER_T(None))']))
struct__EFI_MM_SYSTEM_TABLE._functions_.append(("MmiHandlerUnRegister",['ctypes.c_uint64', 'POINTER_T(None)']))
_EFI_MM_SYSTEM_TABLE = struct__EFI_MM_SYSTEM_TABLE
P_EFI_MM_SYSTEM_TABLE = POINTER_T(struct__EFI_MM_SYSTEM_TABLE)
EFI_MM_SYSTEM_TABLE = struct__EFI_MM_SYSTEM_TABLE
EFI_TABLE_HEADER = struct_EFI_TABLE_HEADER
PEFI_TABLE_HEADER = POINTER_T(struct_EFI_TABLE_HEADER)
CHAR16 = ctypes.c_uint16
UINT32 = ctypes.c_uint32
UINT64 = ctypes.c_uint64
UINTN = ctypes.c_uint64
RETURN_STATUS = ctypes.c_uint64
EFI_STATUS = ctypes.c_uint64
GUID = struct_GUID
PGUID = POINTER_T(struct_GUID)
EFI_MM_INSTALL_CONFIGURATION_TABLE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_SYSTEM_TABLE), POINTER_T(struct_GUID), POINTER_T(None), ctypes.c_uint64))
_EFI_MM_CPU_IO_PROTOCOL = struct__EFI_MM_CPU_IO_PROTOCOL
P_EFI_MM_CPU_IO_PROTOCOL = POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL)
EFI_MEMORY_TYPE = enum_16
EFI_MEMORY_TYPE__enumvalues = enum_16__enumvalues
EFI_ALLOCATE_POOL = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_16, ctypes.c_uint64, POINTER_T(POINTER_T(None))))
EFI_FREE_POOL = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None)))
EFI_ALLOCATE_TYPE = enum_494
EFI_ALLOCATE_TYPE__enumvalues = enum_494__enumvalues
EFI_PHYSICAL_ADDRESS = ctypes.c_uint64
EFI_ALLOCATE_PAGES = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_494, enum_16, ctypes.c_uint64, POINTER_T(ctypes.c_uint64)))
EFI_FREE_PAGES = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, ctypes.c_uint64, ctypes.c_uint64))
EFI_AP_PROCEDURE = POINTER_T(ctypes.CFUNCTYPE(None, POINTER_T(None)))
EFI_MM_STARTUP_THIS_AP = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.CFUNCTYPE(None, POINTER_T(None))), ctypes.c_uint64, POINTER_T(None)))
EFI_CONFIGURATION_TABLE = struct_EFI_CONFIGURATION_TABLE
PEFI_CONFIGURATION_TABLE = POINTER_T(struct_EFI_CONFIGURATION_TABLE)
EFI_HANDLE = POINTER_T(None)
EFI_INTERFACE_TYPE = enum_498
EFI_INTERFACE_TYPE__enumvalues = enum_498__enumvalues
EFI_INSTALL_PROTOCOL_INTERFACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(POINTER_T(None)), POINTER_T(struct_GUID), enum_498, POINTER_T(None)))
EFI_UNINSTALL_PROTOCOL_INTERFACE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID), POINTER_T(None)))
EFI_HANDLE_PROTOCOL = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(struct_GUID), POINTER_T(POINTER_T(None))))
EFI_MM_NOTIFY_FN = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None)))
EFI_MM_REGISTER_PROTOCOL_NOTIFY = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None))), POINTER_T(POINTER_T(None))))
EFI_LOCATE_SEARCH_TYPE = enum_500
EFI_LOCATE_SEARCH_TYPE__enumvalues = enum_500__enumvalues
EFI_LOCATE_HANDLE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, enum_500, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(ctypes.c_uint64), POINTER_T(POINTER_T(None))))
EFI_LOCATE_PROTOCOL = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(POINTER_T(None))))
EFI_MM_INTERRUPT_MANAGE = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct_GUID), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))
EFI_MM_HANDLER_ENTRY_POINT = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64)))
EFI_MM_INTERRUPT_REGISTER = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None), POINTER_T(None), POINTER_T(None), POINTER_T(ctypes.c_uint64))), POINTER_T(struct_GUID), POINTER_T(POINTER_T(None))))
EFI_MM_INTERRUPT_UNREGISTER = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(None)))
UINT16 = ctypes.c_uint16
UINT8 = ctypes.c_ubyte
EFI_MM_IO_ACCESS = struct_EFI_MM_IO_ACCESS
PEFI_MM_IO_ACCESS = POINTER_T(struct_EFI_MM_IO_ACCESS)
EFI_MM_IO_WIDTH = enum_859
EFI_MM_IO_WIDTH__enumvalues = enum_859__enumvalues
EFI_MM_CPU_IO = POINTER_T(ctypes.CFUNCTYPE(ctypes.c_uint64, POINTER_T(struct__EFI_MM_CPU_IO_PROTOCOL), enum_859, ctypes.c_uint64, ctypes.c_uint64, POINTER_T(None)))
__all__ = \
    ['AllHandles', 'AllocateAddress', 'AllocateAnyPages',
    'AllocateMaxAddress', 'ByProtocol', 'ByRegisterNotify', 'CHAR16',
    'EFI_ALLOCATE_PAGES', 'EFI_ALLOCATE_POOL', 'EFI_ALLOCATE_TYPE',
    'EFI_ALLOCATE_TYPE__enumvalues', 'EFI_AP_PROCEDURE',
    'EFI_CONFIGURATION_TABLE', 'EFI_FREE_PAGES', 'EFI_FREE_POOL',
    'EFI_GUID', 'EFI_HANDLE', 'EFI_HANDLE_PROTOCOL',
    'EFI_INSTALL_PROTOCOL_INTERFACE', 'EFI_INTERFACE_TYPE',
    'EFI_INTERFACE_TYPE__enumvalues', 'EFI_LOCATE_HANDLE',
    'EFI_LOCATE_PROTOCOL', 'EFI_LOCATE_SEARCH_TYPE',
    'EFI_LOCATE_SEARCH_TYPE__enumvalues', 'EFI_MEMORY_TYPE',
    'EFI_MEMORY_TYPE__enumvalues', 'EFI_MM_CPU_IO',
    'EFI_MM_CPU_IO_PROTOCOL', 'EFI_MM_HANDLER_ENTRY_POINT',
    'EFI_MM_INSTALL_CONFIGURATION_TABLE', 'EFI_MM_INTERRUPT_MANAGE',
    'EFI_MM_INTERRUPT_REGISTER', 'EFI_MM_INTERRUPT_UNREGISTER',
    'EFI_MM_IO_ACCESS', 'EFI_MM_IO_WIDTH',
    'EFI_MM_IO_WIDTH__enumvalues', 'EFI_MM_NOTIFY_FN',
    'EFI_MM_REGISTER_PROTOCOL_NOTIFY', 'EFI_MM_STARTUP_THIS_AP',
    'EFI_MM_SYSTEM_TABLE', 'EFI_NATIVE_INTERFACE',
    'EFI_PHYSICAL_ADDRESS', 'EFI_STATUS', 'EFI_TABLE_HEADER',
    'EFI_UNINSTALL_PROTOCOL_INTERFACE', 'EfiACPIMemoryNVS',
    'EfiACPIReclaimMemory', 'EfiBootServicesCode',
    'EfiBootServicesData', 'EfiConventionalMemory', 'EfiLoaderCode',
    'EfiLoaderData', 'EfiMaxMemoryType', 'EfiMemoryMappedIO',
    'EfiMemoryMappedIOPortSpace', 'EfiPalCode', 'EfiPersistentMemory',
    'EfiReservedMemoryType', 'EfiRuntimeServicesCode',
    'EfiRuntimeServicesData', 'EfiUnusableMemory', 'GUID',
    'ImageBaseOffset32', 'MM_IO_UINT16', 'MM_IO_UINT32',
    'MM_IO_UINT64', 'MM_IO_UINT8', 'MaxAllocateType',
    'PEFI_CONFIGURATION_TABLE', 'PEFI_MM_IO_ACCESS',
    'PEFI_TABLE_HEADER', 'PGUID', 'P_EFI_MM_CPU_IO_PROTOCOL',
    'P_EFI_MM_SYSTEM_TABLE', 'RETURN_STATUS', 'UINT16', 'UINT32',
    'UINT64', 'UINT8', 'UINTN', '_EFI_MM_CPU_IO_PROTOCOL',
    '_EFI_MM_SYSTEM_TABLE', 'byte', 'dword', 'enum_16', 'enum_494',
    'enum_498', 'enum_500', 'enum_859', 'longlong', 'qword',
    'struct_EFI_CONFIGURATION_TABLE', 'struct_EFI_MM_IO_ACCESS',
    'struct_EFI_TABLE_HEADER', 'struct_GUID',
    'struct__EFI_MM_CPU_IO_PROTOCOL', 'struct__EFI_MM_SYSTEM_TABLE',
    'uchar', 'uint', 'ulonglong', 'undefined', 'undefined1',
    'undefined2', 'undefined4', 'undefined8', 'ushort', 'word']



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 
import struct
from unicorn.x86_const import *
from qiling.os.utils import *


DWORD = 1
UINT = 1
INT = 1
BOOL = 1
SIZE_T = 1
BYTE = 1
ULONGLONG = 2
HANDLE = 3
POINTER = 3
STRING = 4
WSTRING = 5


def read_cstring(ql, address):
    result = ""
    char = ql.mem.read(address, 1)
    while char.decode(errors="ignore") != "\x00":
        address += 1
        result += char.decode(errors="ignore")
        char = ql.mem.read(address, 1)
    return result


def _get_param_by_index(ql, index):
    if ql.archtype == QL_ARCH.X86:
        return _x86_get_params_by_index(ql, index)
    elif ql.archtype == QL_ARCH.X8664:
        return _x8664_get_params_by_index(ql, index)


# TODO: x86 calling convention of Linux kernel?
def _x86_get_params_by_index(ql, index):
    # index starts from 0
    # skip ret_addr
    return ql.stack_read((index + 1) * 4)


def _x86_get_args(ql, number):
    arg_list = []
    for i in range(number):
        # skip ret_addr
        arg_list.append(ql.stack_read((i + 1) * 4))
    if number == 1:
        return arg_list[0]
    else:
        return arg_list


# The kernel interface uses RDI, RSI, RDX, R10, R8 and R9.
def _x8664_get_params_by_index(ql, index):
    reg_list = [UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_RCX, UC_X86_REG_R8, UC_X86_REG_R9]
    if index < 6:
        return ql.uc.reg_read(reg_list[index])

    index -= 6
    # skip ret_addr
    return ql.stack_read((index + 1) * 8)


def _x8664_get_args(ql, number):
    reg_list = [UC_X86_REG_RDI, UC_X86_REG_RSI, UC_X86_REG_RDX, UC_X86_REG_RCX, UC_X86_REG_R8, UC_X86_REG_R9]
    arg_list = []
    reg_num = number
    if reg_num > 6:
        reg_num = 6
    number -= reg_num
    # get args in registers first
    for i in reg_list[:reg_num]:
        arg_list.append(ql.uc.reg_read(i))
    # the rest args are from stack
    for i in range(number):
        # skip ret_addr and 32 byte home space
        arg_list.append(ql.stack_read((i + 7) * 8))
    if reg_num == 1:
        return arg_list[0]
    else:
        return arg_list


def set_function_params(ql, in_params, out_params):
    index = 0
    for each in in_params:
        if in_params[each] == DWORD or in_params[each] == POINTER:
            out_params[each] = _get_param_by_index(ql, index)
        elif in_params[each] == ULONGLONG:
            if ql.archtype == QL_ARCH.X86:
                low = _get_param_by_index(ql, index)
                index += 1
                high = _get_param_by_index(ql, index)
                out_params[each] = high << 8 + low
            else:
                out_params[each] = _get_param_by_index(ql, index)
        elif in_params[each] == STRING:
            ptr = _get_param_by_index(ql, index)
            if ptr == 0:
                out_params[each] = 0
            else:
                out_params[each] = read_cstring(ql, ptr)
        elif in_params[each] == WSTRING:
            ptr = _get_param_by_index(ql, index)
            if ptr == 0:
                out_params[each] = 0
            else:
                out_params[each] = read_wstring(ql, ptr)
        index += 1
    return index


def set_return_value(ql, ret):
    if ql.archtype == QL_ARCH.X86:
        ql.uc.reg_write(UC_X86_REG_EAX, ret)
    elif ql.archtype == QL_ARCH.X8664:
        ql.uc.reg_write(UC_X86_REG_RAX, ret)


def get_return_value(ql):
    if ql.archtype == QL_ARCH.X86:
        return ql.uc.reg_read(UC_X86_REG_EAX)
    elif ql.archtype == QL_ARCH.X8664:
        return ql.uc.reg_read(UC_X86_REG_RAX)


def print_function(ql, passthru, address, function_name, params, ret):
    PRINTK_LEVEL = {
        0: 'KERN_EMERGE',
        1: 'KERN_ALERT',
        1: 'KERN_CRIT',
        2: 'KERN_INFO',
        3: 'KERN_ERR',
        4: 'KERN_WARNING',
        5: 'KERN_NOTICE',
        6: 'KERN_INFO',
        7: 'KERN_DEBUG',
        8: '',
        9: 'KERN_CONT',
    }
    function_name = function_name.replace('hook_', '')
    if function_name in ("__stdio_common_vfprintf", "__stdio_common_vfwprintf",
                         "printf", "wsprintfW", "sprintf"):
        return
    log = '0x%0.2x: %s(' % (address, function_name)
    for each in params:
        value = params[each]
        if type(value) == str or type(value) == bytearray:
            if function_name == 'printk':
                info = value[:2]
                try:
                    level = PRINTK_LEVEL[int(info[1])]
                    value = value[2:]
                    log += '%s = %s "%s", ' %(each, level, value)
                except:
                    log += '%s = "%s", ' %(each, value)
            else:
                log += '%s = "%s", ' %(each, value)
        elif type(value) == tuple:
            log += '%s = 0x%x, ' % (each, value[0])
        else:
            log += '%s = 0x%x, ' % (each, value)
    log = log.strip(", ")
    log += ')'
    if ret is not None:
        # do not print result for printk()
        if function_name != 'printk':
            log += ' = 0x%x' % ret

    if passthru:
        log += ' (PASSTHRU)'

    # replace \n
    log = log.replace("\n", "\\n")
    if ql.output != QL_OUTPUT.DEBUG:
        log = log.partition(" ")[-1]
        ql.nprint(log)
    else:
        ql.dprint(D_INFO, log)


def __x86_cc(ql, param_num, params, func, args, kwargs):
    # read params
    if params is not None:
        param_num = set_function_params(ql, params, args[2])
    # call function
    result = func(*args, **kwargs)
    # set return value
    if result is not None:
        set_return_value(ql, result)
    # print
    print_function(ql, False, args[1], func.__name__, args[2], result)
    return result, param_num


def x86_stdcall(ql, param_num, params, func, args, kwargs):
    # get ret addr
    ret_addr = ql.stack_read(0)

    result, param_num = __x86_cc(ql, param_num, params, func, args, kwargs)

    # update stack pointer
    ql.reg.arch_sp += (param_num + 1) * 4

    ql.reg.arch_pc = ret_addr

    return result


def x86_cdecl(ql, param_num, params, func, args, kwargs):
    result, param_num = __x86_cc(ql, param_num, params, func, args, kwargs)

    ql.reg.arch_pc = ql.stack_pop()

    return result


def x8664_fastcall(ql, param_num, params, func, args, kwargs):
    result, param_num = __x86_cc(ql, param_num, params, func, args, kwargs)

    ql.reg.arch_pc = ql.stack_pop()

    return result


def linux_kernel_api(param_num=None, params=None):
    """
    @cc: windows api calling convention, only x86 needs this, x64 is always fastcall
    @params: params dict
    @param_num: the number of function params, used by variadic functions, e.g printf
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            ql = args[0]
            if ql.archtype == QL_ARCH.X86:
                # if cc == STDCALL:
                return x86_stdcall(ql, param_num, params, func, args, kwargs)
                #elif cc == CDECL:
                #    return x86_cdecl(ql, param_num, params, func, args, kwargs)
            elif ql.archtype == QL_ARCH.X8664:
                return x8664_fastcall(ql, param_num, params, func, args, kwargs)
            else:
                raise QlErrorArch("[!] Unknown ql.archtype")
        return wrapper
    return decorator




from qiling.const import *
from qiling.os.const import *
from .const import *
from .utils import *
from .mm_access_type import *
from .fncc import *


@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)
})
def hook_Open(ql, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)
})
def hook_Close(ql, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)
})
def hook_Lock(ql, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "This": POINTER, #POINTER_T(struct__EFI_MM_ACCESS_PROTOCOL)
    "MmramMapSize": POINTER, #POINTER_T(ctypes.c_uint64)
    "MmramMap": POINTER, #POINTER_T(struct_EFI_MMRAM_DESCRIPTOR)
})
def hook_GetCapabilities(ql, address, params):
    write_int64(ql, params["MmramMapSize"], 0)
    if params['MmramMap'] != 0:
        write_int64(ql, params["MmramMap"], 0)
    return EFI_SUCCESS



def install_EFI_MM_ACCESS_PROTOCOL(ql, start_ptr):
    efi_mm_access_protocol = EFI_MM_ACCESS_PROTOCOL()
    ptr = start_ptr
    pointer_size = 8
    efi_mm_access_protocol.Open = ptr
    ql.hook_address(hook_Open, ptr)
    ptr += pointer_size
    efi_mm_access_protocol.Close = ptr
    ql.hook_address(hook_Close, ptr)
    ptr += pointer_size
    efi_mm_access_protocol.Lock = ptr
    ql.hook_address(hook_Lock, ptr)
    ptr += pointer_size
    efi_mm_access_protocol.GetCapabilities = ptr
    ql.hook_address(hook_GetCapabilities, ptr)
    ptr += pointer_size
    return (ptr, efi_mm_access_protocol)




#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import ctypes, types, struct, ast
from contextlib import contextmanager

from unicorn import *
from unicorn.x86_const import *

from qiling.arch.x86_const import *
from qiling.os.utils import *
from qiling.const import *
from qiling.os.memory import QlMemoryHeap
from qiling.os.uefi.utils import *
from qiling.os.uefi.type64 import *
from qiling.os.uefi.fncc import *
from qiling.os.uefi.bootup import *
from qiling.os.uefi.runtime import *
from qiling.os.uefi.dxe_service import *
from qiling.os.uefi.smm_base2_protocol import *
from qiling.os.uefi.mm_access_protocol import *
from qiling.os.uefi.smm_sw_dispatch2_protocol import *

from qiling.os.windows.fncc import *

import pefile
from .loader import QlLoader

class QlLoaderPE_UEFI(QlLoader):
    def __init__(self, ql):
        super(QlLoaderPE_UEFI, self).__init__(ql)
        self.ql = ql
        self.modules = []
        self.events = {}
        self.handle_dict = {}
        self.notify_list = []
        self.next_image_base = 0x10000

    def save(self):
        saved_state = super(QlLoaderPE_UEFI, self).save()
        
        # We can't serialize self.modules since it contain pefile objects. let's remove it now and generate it again when loading.
        modules = []
        for mod in self.modules:
            modules.append(mod[:3])
        saved_state['modules'] = modules

        saved_state['events'] = self.events
        saved_state['handle_dict'] = self.handle_dict
        saved_state['notify_list'] = self.notify_list
        saved_state['next_image_base'] = self.next_image_base
        saved_state['loaded_image_protocol_modules'] = self.loaded_image_protocol_modules
        saved_state['tpl'] = self.tpl
        saved_state['efi_configuration_table'] = self.efi_configuration_table
        # since this class initialize the heap (that is hosted by the OS object), we will store it here.
        saved_state['heap'] = self.ql.os.heap.save()
        return saved_state

    def restore(self, saved_state):
        super(QlLoaderPE_UEFI, self).restore(saved_state)
        self.modules = []
        for mod in saved_state['modules']:
            self.modules.append(mod+(pefile.PE(mod[0], fast_load=True),))
        self.events = saved_state['events']
        self.handle_dict = saved_state['handle_dict']
        self.notify_list = saved_state['notify_list']
        self.next_image_base = saved_state['next_image_base']
        self.loaded_image_protocol_modules = saved_state['loaded_image_protocol_modules']
        self.tpl = saved_state['tpl']
        self.efi_configuration_table = saved_state['efi_configuration_table']
        self.ql.os.heap.restore(saved_state['heap'])

    @contextmanager
    def map_memory(self, addr, size):
        self.ql.mem.map(addr, size)
        try:
            yield
        finally:
            self.ql.mem.unmap(addr, size)


    def install_loaded_image_protocol(self, image_base, image_size, entry_point):
        loaded_image_protocol = EFI_LOADED_IMAGE_PROTOCOL()
        loaded_image_protocol.Revision = int(self.ql.os.profile["LOADED_IMAGE_PROTOCOL"]["revision"], 16)
        loaded_image_protocol.ParentHandle = 0
        loaded_image_protocol.SystemTable = self.system_table_ptr
        loaded_image_protocol.DeviceHandle = image_base
        loaded_image_protocol.FilePath = 0 # This is a handle to a complex path object, skip it for now.
        loaded_image_protocol.LoadOptionsSize = 0
        loaded_image_protocol.LoadOptions = 0
        loaded_image_protocol.ImageBase = image_base
        loaded_image_protocol.ImageSize = image_size
        loaded_image_protocol.ImageCodeType = EfiLoaderCode
        loaded_image_protocol.ImageDataType = EfiLoaderData
        loaded_image_protocol.Unload = 0

        loaded_image_protocol_ptr = self.ql.os.heap.alloc(ctypes.sizeof(EFI_LOADED_IMAGE_PROTOCOL))
        self.ql.mem.write(loaded_image_protocol_ptr, convert_struct_to_bytes(loaded_image_protocol))
        self.handle_dict[image_base] = {self.loaded_image_protocol_guid: loaded_image_protocol_ptr}
        self.loaded_image_protocol_modules.append(image_base)


    def map_and_load(self, path, execute_now=False):
        ql = self.ql
        pe = pefile.PE(path, fast_load=True)

        # Make sure no module will occupy the NULL page
        if self.next_image_base > pe.OPTIONAL_HEADER.ImageBase:
            IMAGE_BASE = self.next_image_base
            pe.relocate_image(IMAGE_BASE)
        else:
            IMAGE_BASE = pe.OPTIONAL_HEADER.ImageBase
        IMAGE_SIZE = ql.mem.align(pe.OPTIONAL_HEADER.SizeOfImage, 0x1000)

        while IMAGE_BASE + IMAGE_SIZE < self.heap_base_address:
            if not ql.mem.is_mapped(IMAGE_BASE, 1):
                self.next_image_base = IMAGE_BASE + 0x10000
                ql.mem.map(IMAGE_BASE, IMAGE_SIZE)
                pe.parse_data_directories()
                data = bytearray(pe.get_memory_mapped_image())
                ql.mem.write(IMAGE_BASE, bytes(data))
                ql.nprint("[+] Loading %s to 0x%x" % (path, IMAGE_BASE))
                entry_point = IMAGE_BASE + pe.OPTIONAL_HEADER.AddressOfEntryPoint
                if self.entry_point == 0:
                    # Setting entry point to the first loaded module entry point, so the debugger can break.
                    self.entry_point = entry_point
                ql.nprint("[+] PE entry point at 0x%x" % entry_point)
                self.install_loaded_image_protocol(IMAGE_BASE, IMAGE_SIZE, entry_point)
                self.images.append(self.coverage_image(IMAGE_BASE, IMAGE_BASE + pe.NT_HEADERS.OPTIONAL_HEADER.SizeOfImage, path))
                if execute_now:
                    self.ql.nprint(f'[+] Running from 0x{entry_point:x} of {path}')
                    assembler = self.ql.create_assembler()
                    code = f"""
                        mov rcx, {IMAGE_BASE}
                        mov rdx, {self.system_table_ptr}
                        mov rax, {entry_point}
                        call rax
                    """
                    runcode, _ = assembler.asm(code)
                    ptr = ql.os.heap.alloc(len(runcode))
                    ql.mem.write(ptr, bytes(runcode))
                    ql.os.exec_arbitrary(ptr, ptr+len(runcode))

                else:
                    self.modules.append((path, IMAGE_BASE, entry_point, pe))
                return True
            else:
                IMAGE_BASE += 0x10000
                pe.relocate_image(IMAGE_BASE)
        return False

    def unload_modules(self):
        for handle in self.loaded_image_protocol_modules:
            dic = self.handle_dict[handle]
            buf = bytes(self.ql.mem.read(dic[self.loaded_image_protocol_guid], ctypes.sizeof(EFI_LOADED_IMAGE_PROTOCOL)))
            buffer = ctypes.create_string_buffer(buf)
            loaded_image_protocol = EFI_LOADED_IMAGE_PROTOCOL()
            ctypes.memmove(ctypes.addressof(loaded_image_protocol), buffer, ctypes.sizeof(loaded_image_protocol))
            unload_ptr = struct.unpack("Q", loaded_image_protocol.Unload)[0]
            if unload_ptr != 0:
                self.ql.stack_push(self.end_of_execution_ptr)
                self.ql.reg.rcx = handle
                self.ql.reg.rip = unload_ptr
                self.ql.nprint(f'[+] Unloading module 0x{handle:x}, calling 0x{unload_ptr:x}')
                self.loaded_image_protocol_modules.remove(handle)
                return True
        return False

    def execute_module(self, path, image_base, entry_point, EOE_ptr):
        self.ql.stack_push(EOE_ptr)
        self.ql.reg.rcx = image_base
        self.ql.reg.rdx = self.system_table_ptr
        self.ql.reg.rip = entry_point
        self.ql.os.entry_point = entry_point
        self.ql.nprint(f'[+] Running from 0x{entry_point:x} of {path}')

    def execute_next_module(self):
        if self.ql.os.notify_before_module_execution(self.ql, self.modules[0][0]):
            return
        path, image_base, entry_point, pe = self.modules.pop(0)
        self.execute_module(path, image_base, entry_point, self.end_of_execution_ptr)


    def run(self):
        self.loaded_image_protocol_guid = self.ql.os.profile["LOADED_IMAGE_PROTOCOL"]["guid"]
        self.loaded_image_protocol_modules = []
        self.tpl = 4 # TPL_APPLICATION
        self.user_defined_api = self.ql.os.user_defined_api
        self.user_defined_api_onenter = self.ql.os.user_defined_api_onenter
        self.user_defined_api_onexit = self.ql.os.user_defined_api_onexit
        
        if self.ql.archtype == QL_ARCH.X8664:
            self.heap_base_address = int(self.ql.os.profile.get("OS64", "heap_address"), 16)
            self.heap_base_size = int(self.ql.os.profile.get("OS64", "heap_size"), 16)       
        elif self.ql.archtype == QL_ARCH.X86:
            self.heap_base_address = int(self.ql.os.profile.get("OS32", "heap_address"), 16)
            self.heap_base_size = int(self.ql.os.profile.get("OS32", "heap_size"), 16)
        
        self.ql.os.heap = QlMemoryHeap(self.ql, self.heap_base_address, self.heap_base_address + self.heap_base_size)
        self.entry_point = 0
        self.load_address = 0  

        if self.ql.archtype == QL_ARCH.X8664:
            self.stack_address = int(self.ql.os.profile.get("OS64", "stack_address"), 16)
            self.stack_size = int(self.ql.os.profile.get("OS64", "stack_size"), 16)
            
        elif self.ql.archtype == QL_ARCH.X86:        
            self.stack_address = int(self.ql.os.profile.get("OS32", "stack_address"), 16)
            self.stack_size = int(self.ql.os.profile.get("OS32", "stack_size"), 16)     

        # set stack pointer
        self.ql.nprint("[+] Initiate stack address at 0x%x" % self.stack_address)
        self.ql.mem.map(self.stack_address, self.stack_size)

        # Stack should not init at the very bottom. Will cause errors with Dlls
        sp = self.stack_address + self.stack_size - 0x1000

        if self.ql.archtype== QL_ARCH.X86:
            raise QlErrorArch("[!] Only 64 bit arch supported for now.")

        elif self.ql.archtype== QL_ARCH.X8664:
            self.ql.reg.rsp = sp
            self.ql.reg.rbp = sp

        else:
            raise QlErrorArch("[!] Unknown ql.arch")

        # set SystemTable to image base for now
        pointer_size = ctypes.sizeof(ctypes.c_void_p)
        system_table_heap_size = 1024 * 1024
        system_table_heap = self.ql.os.heap.alloc(system_table_heap_size)
        self.ql.mem.write(system_table_heap, b'\x90'*system_table_heap_size)
        self.system_table_ptr = system_table_heap
        system_table = EFI_SYSTEM_TABLE()
        system_table_heap_ptr = system_table_heap + ctypes.sizeof(EFI_SYSTEM_TABLE)
        
        self.runtime_services_ptr = system_table_heap_ptr
        system_table.RuntimeServices = self.runtime_services_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_RUNTIME_SERVICES)
        system_table_heap_ptr, self.runtime_services = hook_EFI_RUNTIME_SERVICES(self.ql, system_table_heap_ptr)
        self.runtime_services_end_ptr = system_table_heap_ptr - pointer_size

        self.boot_services_ptr = system_table_heap_ptr
        system_table.BootServices = self.boot_services_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_BOOT_SERVICES)
        system_table_heap_ptr, boot_services = hook_EFI_BOOT_SERVICES(self.ql, system_table_heap_ptr)
        self.boot_services_end_ptr = system_table_heap_ptr - pointer_size

        self.efi_configuration_table_ptr = system_table_heap_ptr
        system_table.ConfigurationTable = self.efi_configuration_table_ptr
        system_table.NumberOfTableEntries = 2
        system_table_heap_ptr += ctypes.sizeof(EFI_CONFIGURATION_TABLE) * 100 # We don't expect more then a few entries.
        efi_configuration_table = EFI_CONFIGURATION_TABLE()

        #   0x7739f24c, 0x93d7, 0x11d4, {0x9a, 0x3a, 0x0, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
        efi_configuration_table.VendorGuid.Data1 = int(self.ql.os.profile.get("HOB_LIST", "data1"), 16)
        efi_configuration_table.VendorGuid.Data2 = int(self.ql.os.profile.get("HOB_LIST", "data2"), 16)
        efi_configuration_table.VendorGuid.Data3 = int(self.ql.os.profile.get("HOB_LIST", "data3"), 16)
        
        data4 = ast.literal_eval(self.ql.os.profile.get("HOB_LIST", "data4"))
        datalist = 0
        for data4_list in data4:
            efi_configuration_table.VendorGuid.Data4[datalist] = data4_list
            datalist += 1  
        
        VendorTable_ptr = system_table_heap_ptr
        write_int64(self.ql, VendorTable_ptr, int(self.ql.os.profile.get("HOB_LIST", "vendortable"),16))
        system_table_heap_ptr += pointer_size
        efi_configuration_table.VendorTable = VendorTable_ptr
        self.efi_configuration_table = [self.ql.os.profile["HOB_LIST"]["guid"]]
        self.ql.mem.write(self.efi_configuration_table_ptr, convert_struct_to_bytes(efi_configuration_table))

        self.mm_system_table_ptr = system_table_heap_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_MM_SYSTEM_TABLE)
        system_table_heap_ptr, efi_mm_system_table = create_EFI_MM_SYSTEM_TABLE(self.ql, system_table_heap_ptr)
        self.mm_system_table_end_ptr = system_table_heap_ptr - pointer_size
        efi_mm_system_table.MmConfigurationTable = self.efi_configuration_table_ptr

        self.smm_base2_protocol_ptr = system_table_heap_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_SMM_BASE2_PROTOCOL)
        system_table_heap_ptr, smm_base2_protocol, efi_mm_system_table = install_EFI_SMM_BASE2_PROTOCOL(self.ql, system_table_heap_ptr, efi_mm_system_table)
        self.handle_dict[1] = {self.ql.os.profile.get("EFI_SMM_BASE2_PROTOCOL", "guid"): self.smm_base2_protocol_ptr}

        self.mm_access_protocol_ptr = system_table_heap_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_MM_ACCESS_PROTOCOL)
        system_table_heap_ptr, mm_access_protocol = install_EFI_MM_ACCESS_PROTOCOL(self.ql, system_table_heap_ptr)
        self.handle_dict[1][self.ql.os.profile.get("EFI_MM_ACCESS_PROTOCOL", "guid")] = self.mm_access_protocol_ptr

        self.smm_sw_dispatch2_protocol_ptr = system_table_heap_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_SMM_SW_DISPATCH2_PROTOCOL)
        system_table_heap_ptr, smm_sw_dispatch2_protocol = install_EFI_SMM_SW_DISPATCH2_PROTOCOL(self.ql, system_table_heap_ptr)
        self.handle_dict[1][self.ql.os.profile.get("EFI_SMM_SW_DISPATCH2_PROTOCOL", "guid")] = self.smm_sw_dispatch2_protocol_ptr

        self.dxe_services_ptr = system_table_heap_ptr
        system_table_heap_ptr += ctypes.sizeof(EFI_DXE_SERVICES)
        system_table_heap_ptr, dxe_services = install_EFI_DXE_SERVICES(self.ql, system_table_heap_ptr)
        efi_configuration_table = EFI_CONFIGURATION_TABLE()
        efi_configuration_table.VendorGuid.Data1 = int(self.ql.os.profile.get("DXE_SERVICE_TABLE", "data1"), 16)
        efi_configuration_table.VendorGuid.Data2 = int(self.ql.os.profile.get("DXE_SERVICE_TABLE", "data2"), 16)
        efi_configuration_table.VendorGuid.Data3 = int(self.ql.os.profile.get("DXE_SERVICE_TABLE", "data3"), 16)
        
        data4 = ast.literal_eval(self.ql.os.profile.get("DXE_SERVICE_TABLE", "data4"))
        datalist = 0
        for data4_list in data4:
            efi_configuration_table.VendorGuid.Data4[datalist] = data4_list
            datalist += 1  
       
        efi_configuration_table.VendorTable = self.dxe_services_ptr
        self.ql.mem.write(self.efi_configuration_table_ptr + ctypes.sizeof(EFI_CONFIGURATION_TABLE), convert_struct_to_bytes(efi_configuration_table))
        self.efi_configuration_table.append(self.ql.os.profile.get("DXE_SERVICE_TABLE", "guid"))
        

        self.ql.mem.write(self.runtime_services_ptr, convert_struct_to_bytes(self.runtime_services))
        self.ql.mem.write(self.boot_services_ptr, convert_struct_to_bytes(boot_services))
        self.ql.mem.write(self.system_table_ptr, convert_struct_to_bytes(system_table))
        self.ql.mem.write(self.mm_system_table_ptr, convert_struct_to_bytes(efi_mm_system_table))
        self.ql.mem.write(self.smm_base2_protocol_ptr, convert_struct_to_bytes(smm_base2_protocol))
        self.ql.mem.write(self.mm_access_protocol_ptr, convert_struct_to_bytes(mm_access_protocol))
        self.ql.mem.write(self.smm_sw_dispatch2_protocol_ptr, convert_struct_to_bytes(smm_sw_dispatch2_protocol))
        self.ql.mem.write(self.dxe_services_ptr, convert_struct_to_bytes(dxe_services))

        for dependency in self.ql.argv:
            if not self.map_and_load(dependency):
                raise QlErrorFileType("Can't map dependency")

        self.ql.nprint("[+] Done with loading %s" % self.ql.path)

        #return address
        self.end_of_execution_ptr = system_table_heap_ptr
        self.ql.mem.write(self.end_of_execution_ptr, b'\xcc')
        system_table_heap_ptr += pointer_size
        self.ql.hook_address(hook_EndOfExecution, self.end_of_execution_ptr)

        self.execute_next_module()

    def restore_runtime_services(self):
        self.ql.mem.write(self.runtime_services_ptr, convert_struct_to_bytes(self.runtime_services))




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

import ctypes
import struct
from qiling.const import *
from .const import *

def convert_struct_to_bytes(st):
    buffer = ctypes.create_string_buffer(ctypes.sizeof(st))
    ctypes.memmove(buffer, ctypes.addressof(st), ctypes.sizeof(st))
    return buffer.raw

def check_and_notify_protocols(ql):
    if len(ql.loader.notify_list) > 0:
        event_id, notify_func, notify_context = ql.loader.notify_list.pop(0)
        ql.nprint(f'Notify event:{event_id} calling:{notify_func:x} context:{notify_context:x}')
        ql.stack_push(ql.loader.end_of_execution_ptr)
        ql.reg.rcx = notify_context
        ql.reg.arch_pc = notify_func
        return True
    return False

def write_int32(ql, address, num):
    if ql.archendian == QL_ENDIAN.EL:
        ql.mem.write(address, struct.pack('<I',(num)))
    else:
        ql.mem.write(address, struct.pack('>I',(num)))

def write_int64(ql, address, num):
    if ql.archendian == QL_ENDIAN.EL:
        ql.mem.write(address, struct.pack('<Q',(num)))
    else:
        ql.mem.write(address, struct.pack('>Q',(num)))

def read_int64(ql, address):
    if ql.archendian == QL_ENDIAN.EL:
        return struct.unpack('<Q', ql.mem.read(address, 8))[0]
    else:
        return struct.unpack('>Q',ql.mem.read(address, 8))[0]

def LocateHandles(ql, address, params):
    handles = []
    if params["SearchType"] == SEARCHTYPE_AllHandles:
        handles = ql.loader.handle_dict.keys()
    elif params["SearchType"] == SEARCHTYPE_ByProtocol:
        for handle, guid_dic in ql.loader.handle_dict.items():
            if params["Protocol"] in guid_dic:
                handles.append(handle)
                    
    return len(handles) * pointer_size, handles
    
def LocateProtocol(ql, address, params):
    protocol = params['Protocol']
    for handle, guid_dic in ql.loader.handle_dict.items():
        if "Handle" in params and params["Handle"] != handle:
            continue
        if protocol in guid_dic:
            write_int64(ql, params['Interface'], guid_dic[protocol])
            return EFI_SUCCESS
    return EFI_NOT_FOUND


#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

"""
This module is intended for general purpose functions that are only used in qiling.arch
"""


from qiling.const import *
from qiling.os.const import *
from .const import *
from .utils import *
from .dxe_service_type64 import *
from .fncc import *

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": ULONGLONG,
})
def hook_AddMemorySpace(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": ULONGLONG,
    "a4": POINTER, #POINTER_T(ctypes.c_uint64)
    "a5": POINTER, #POINTER_T(None)
    "a6": POINTER, #POINTER_T(None)
})
def hook_AllocateMemorySpace(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
})
def hook_FreeMemorySpace(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
})
def hook_RemoveMemorySpace(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR)
})
def hook_GetMemorySpaceDescriptor(ctx, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
})
def hook_SetMemorySpaceAttributes(ctx, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "a0": POINTER, #POINTER_T(ctypes.c_uint64)
    "a1": POINTER, #POINTER_T(POINTER_T(struct_EFI_GCD_MEMORY_SPACE_DESCRIPTOR))
})
def hook_GetMemorySpaceMap(ctx, address, params):
    return EFI_UNSUPPORTED

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
})
def hook_AddIoSpace(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": ULONGLONG,
    "a4": POINTER, #POINTER_T(ctypes.c_uint64)
    "a5": POINTER, #POINTER_T(None)
    "a6": POINTER, #POINTER_T(None)
})
def hook_AllocateIoSpace(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
})
def hook_FreeIoSpace(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
})
def hook_RemoveIoSpace(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR)
})
def hook_GetIoSpaceDescriptor(ctx, address, params):
    return EFI_NOT_FOUND

@dxeapi(params={
    "a0": POINTER, #POINTER_T(ctypes.c_uint64)
    "a1": POINTER, #POINTER_T(POINTER_T(struct_EFI_GCD_IO_SPACE_DESCRIPTOR))
})
def hook_GetIoSpaceMap(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
})
def hook_Dispatch(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": GUID,
})
def hook_Schedule(ctx, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": GUID,
})
def hook_Trust(ctx, address, params):
    return EFI_NOT_FOUND

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": ULONGLONG,
    "a2": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_ProcessFirmwareVolume(ctx, address, params):
    return EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
})
def hook_SetMemorySpaceCapabilities(ctx, address, params):
    return EFI_UNSUPPORTED



def install_EFI_DXE_SERVICES(ql, start_ptr):
    efi_dxe_services = EFI_DXE_SERVICES()
    ptr = start_ptr
    pointer_size = 8
    efi_dxe_services.AddMemorySpace = ptr
    ql.hook_address(hook_AddMemorySpace, ptr)
    ptr += pointer_size
    efi_dxe_services.AllocateMemorySpace = ptr
    ql.hook_address(hook_AllocateMemorySpace, ptr)
    ptr += pointer_size
    efi_dxe_services.FreeMemorySpace = ptr
    ql.hook_address(hook_FreeMemorySpace, ptr)
    ptr += pointer_size
    efi_dxe_services.RemoveMemorySpace = ptr
    ql.hook_address(hook_RemoveMemorySpace, ptr)
    ptr += pointer_size
    efi_dxe_services.GetMemorySpaceDescriptor = ptr
    ql.hook_address(hook_GetMemorySpaceDescriptor, ptr)
    ptr += pointer_size
    efi_dxe_services.SetMemorySpaceAttributes = ptr
    ql.hook_address(hook_SetMemorySpaceAttributes, ptr)
    ptr += pointer_size
    efi_dxe_services.GetMemorySpaceMap = ptr
    ql.hook_address(hook_GetMemorySpaceMap, ptr)
    ptr += pointer_size
    efi_dxe_services.AddIoSpace = ptr
    ql.hook_address(hook_AddIoSpace, ptr)
    ptr += pointer_size
    efi_dxe_services.AllocateIoSpace = ptr
    ql.hook_address(hook_AllocateIoSpace, ptr)
    ptr += pointer_size
    efi_dxe_services.FreeIoSpace = ptr
    ql.hook_address(hook_FreeIoSpace, ptr)
    ptr += pointer_size
    efi_dxe_services.RemoveIoSpace = ptr
    ql.hook_address(hook_RemoveIoSpace, ptr)
    ptr += pointer_size
    efi_dxe_services.GetIoSpaceDescriptor = ptr
    ql.hook_address(hook_GetIoSpaceDescriptor, ptr)
    ptr += pointer_size
    efi_dxe_services.GetIoSpaceMap = ptr
    ql.hook_address(hook_GetIoSpaceMap, ptr)
    ptr += pointer_size
    efi_dxe_services.Dispatch = ptr
    ql.hook_address(hook_Dispatch, ptr)
    ptr += pointer_size
    efi_dxe_services.Schedule = ptr
    ql.hook_address(hook_Schedule, ptr)
    ptr += pointer_size
    efi_dxe_services.Trust = ptr
    ql.hook_address(hook_Trust, ptr)
    ptr += pointer_size
    efi_dxe_services.ProcessFirmwareVolume = ptr
    ql.hook_address(hook_ProcessFirmwareVolume, ptr)
    ptr += pointer_size
    efi_dxe_services.SetMemorySpaceCapabilities = ptr
    ql.hook_address(hook_SetMemorySpaceCapabilities, ptr)
    ptr += pointer_size
    return (ptr, efi_dxe_services)




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

import os, logging, configparser

from keystone import *
from capstone import *
from binascii import unhexlify

from .utils import ql_build_module_import_name, ql_get_module_function
from .utils import ql_is_valid_arch, ql_is_valid_ostype
from .utils import loadertype_convert_str, ostype_convert_str, arch_convert_str
from .utils import ql_setup_filter, debugger_convert
from .const import QL_OS, QL_OS_ALL, QL_ARCH, QL_ENDIAN, QL_OUTPUT, QL_DEBUGGER
from .const import D_INFO, D_DRPT
from .exception import QlErrorArch, QlErrorOsType, QlErrorOutput
from .loader.utils import ql_checkostype


class QlCoreUtils(object):
    def __init__(self):
        super().__init__()
        self.archtype = None
        self.ostype = None
        self.path = None
        self.archendian = None

    # normal print out
    def nprint(self, *args, **kw):
        if type(self.console) is bool:
            pass
        else:
            raise QlErrorOutput("[!] console must be True or False")     
        
        # FIXME: this is due to console must be able to update during runtime
        if self.log_file_fd is not None:
            if self.multithread == True and self.os.thread_management is not None and self.os.thread_management.cur_thread is not None:
                fd = self.os.thread_management.cur_thread.log_file_fd
            else:
                fd = self.log_file_fd

            # setup filter for logger
            # FIXME: only works for logging due to we might need runtime disable nprint, it should be a global filter not only syscall
            if self.filter != None and self.output == QL_OUTPUT.DEFAULT:
                fd.addFilter(ql_setup_filter(self.filter))
            
            console_handlers = []

            for each_handler in fd.handlers:
                if type(each_handler) == logging.StreamHandler:
                    console_handlers.append(each_handler)

            if self.console == False:
                for each_console_handler in console_handlers:
                    if '_FalseFilter' not in [each.__class__.__name__ for each in each_console_handler.filters]:
                        each_console_handler.addFilter(ql_setup_filter(False))

            elif self.console == True:
                for each_console_handler in console_handlers:
                    for each_filter in [each for each in each_console_handler.filters]:
                        if '_FalseFilter' in each_filter.__class__.__name__:
                            each_console_handler.removeFilter(each_filter)
            
            args = map(str, args)
            msg = kw.get("sep", " ").join(args)

            if kw.get("end", None) != None:
                msg += kw["end"]

            elif msg != os.linesep:
                msg += os.linesep

            fd.info(msg)


    # debug print out, always use with verbose level with dprint(D_INFO,"helloworld")
    def dprint(self, level, *args, **kw):
        try:
            self.verbose = int(self.verbose)
        except:
            raise QlErrorOutput("[!] Verbose muse be int")    
        
        if type(self.verbose) != int or self.verbose > 99 or (self.verbose > 1 and self.output not in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP)):
            raise QlErrorOutput("[!] Verbose > 1 must use with QL_OUTPUT.DEBUG or else ql.verbose must be 0")

        if self.output == QL_OUTPUT.DUMP:
            self.verbose = 99

        if int(self.verbose) >= level and self.output in (QL_OUTPUT.DEBUG, QL_OUTPUT.DUMP):
            if int(self.verbose) >= D_DRPT:
                try:
                    current_pc = self.reg.arch_pc
                except:
                    current_pc = 0    

                args = (("0x%x:" % current_pc), *args)        
                
            self.nprint(*args, **kw)


    def add_fs_mapper(self, ql_path, real_dest):
        self.os.fs_mapper.add_fs_mapping(ql_path, real_dest)


    # push to stack bottom, and update stack register
    def stack_push(self, data):
        self.arch.stack_push(data)


    # pop from stack bottom, and update stack register
    def stack_pop(self):
        return self.arch.stack_pop()


    # read from stack, at a given offset from stack bottom
    # NOTE: unlike stack_pop(), this does not change stack register
    def stack_read(self, offset):
        return self.arch.stack_read(offset)


    # write to stack, at a given offset from stack bottom
    # NOTE: unlike stack_push(), this does not change stack register
    def stack_write(self, offset, data):
        self.arch.stack_write(offset, data)


    def debugger_setup(self):
        # default remote server
        remotedebugsrv = "gdb"
        debug_opts = [None, None]

        if self.debugger != True and type(self.debugger) == str:      
            debug_opts = self.debugger.split(":")
    
            if len(debug_opts) == 2 and debug_opts[0] != "qdb":
                pass
            else:  
                remotedebugsrv, *debug_opts = debug_opts
                
            
            if debugger_convert(remotedebugsrv) not in (QL_DEBUGGER):
                raise QlErrorOutput("[!] Error: Debugger not supported")
            
        debugsession = ql_get_module_function("qiling.debugger." + remotedebugsrv + "." + remotedebugsrv, "Ql" + str.capitalize(remotedebugsrv))

        return debugsession(self, *debug_opts)

    def arch_setup(self):
        if not ql_is_valid_arch(self.archtype):
            raise QlErrorArch("[!] Invalid Arch")
        
        if self.archtype == QL_ARCH.ARM_THUMB:
            archtype =  QL_ARCH.ARM
        else:
            archtype = self.archtype

        archmanager = arch_convert_str(archtype).upper()
        archmanager = ("QlArch" + archmanager)

        module_name = ql_build_module_import_name("arch", None, archtype)
        return ql_get_module_function(module_name, archmanager)(self)


    def os_setup(self, function_name = None):
        if not ql_is_valid_ostype(self.ostype):
            raise QlErrorOsType("[!] Invalid OSType")

        if not ql_is_valid_arch(self.archtype):
            raise QlErrorArch("[!] Invalid Arch %s" % self.archtype)

        if function_name == None:
            ostype_str = ostype_convert_str(self.ostype)
            ostype_str = ostype_str.capitalize()
            function_name = "QlOs" + ostype_str
            module_name = ql_build_module_import_name("os", self.ostype)
            return ql_get_module_function(module_name, function_name)(self)

        elif function_name == "map_syscall":
            ostype_str = ostype_convert_str(self.ostype)
            arch_str = arch_convert_str(self.archtype)

            syscall_table = "map_syscall"

            module_name = ql_build_module_import_name("os", ostype_str, syscall_table)
            return ql_get_module_function(module_name, function_name)
        
        else:
            module_name = ql_build_module_import_name("os", self.ostype, self.archtype)
            return ql_get_module_function(module_name, function_name)


    def loader_setup(self, function_name = None):
        if not self.shellcoder:
            archtype, ostype, archendian = ql_checkostype(self.path)
            if self.archtype is None:
                self.archtype = archtype
            if self.ostype is None:
                self.ostype = ostype
            if self.archendian is None:
                self.archendian = archendian

        if not ql_is_valid_ostype(self.ostype):
            raise QlErrorOsType("[!] Invalid OSType")

        if not ql_is_valid_arch(self.archtype):
            raise QlErrorArch("[!] Invalid Arch %s" % self.archtype)

        if function_name == None:
            loadertype_str = loadertype_convert_str(self.ostype)
            function_name = "QlLoader" + loadertype_str
            module_name = ql_build_module_import_name("loader", loadertype_str.lower())
            return ql_get_module_function(module_name, function_name)(self)


    def component_setup(self, component_type, function_name):
        if not ql_is_valid_ostype(self.ostype):
            raise QlErrorOsType("[!] Invalid OSType")

        if not ql_is_valid_arch(self.archtype):
            raise QlErrorArch("[!] Invalid Arch %s" % self.archtype)

        module_name = "qiling." + component_type + "." + function_name
        function_name = "Ql" + function_name.capitalize() + "Manager"
        return ql_get_module_function(module_name, function_name)(self)


    def profile_setup(self):
        if self.profile:
            self.dprint(D_INFO, "[+] Customized profile: %s" % self.profile)

        os_profile = os.path.join(os.path.dirname(os.path.abspath(__file__)), "profiles", ostype_convert_str(self.ostype) + ".ql")

        if self.profile:
            profiles = [os_profile, self.profile]
        else:
            profiles = [os_profile]

        config = configparser.ConfigParser()
        config.read(profiles)
        return config
    
    # Assembler/Disassembler API

    def create_disassembler(self):
        if self.archtype == QL_ARCH.ARM:  # QL_ARM
            reg_cpsr = self.reg.cpsr
            mode = CS_MODE_ARM
            if self.archendian == QL_ENDIAN.EB:
                reg_cpsr_v = 0b100000
                # reg_cpsr_v = 0b000000
            else:
                reg_cpsr_v = 0b100000

            if reg_cpsr & reg_cpsr_v != 0:
                mode = CS_MODE_THUMB

            if self.archendian == QL_ENDIAN.EB:
                md = Cs(CS_ARCH_ARM, mode)
                # md = Cs(CS_ARCH_ARM, mode + CS_MODE_BIG_ENDIAN)
            else:
                md = Cs(CS_ARCH_ARM, mode)

        elif self.archtype == QL_ARCH.ARM_THUMB:
            md = Cs(CS_ARCH_ARM, CS_MODE_THUMB)

        elif self.archtype == QL_ARCH.X86:  # QL_X86
            md = Cs(CS_ARCH_X86, CS_MODE_32)

        elif self.archtype == QL_ARCH.X8664:  # QL_X86_64
            md = Cs(CS_ARCH_X86, CS_MODE_64)

        elif self.archtype == QL_ARCH.ARM64:  # QL_ARM64
            md = Cs(CS_ARCH_ARM64, CS_MODE_ARM)

        elif self.archtype == QL_ARCH.A8086:  # QL_A8086
            md = Cs(CS_ARCH_X86, CS_MODE_16)

        elif self.archtype == QL_ARCH.MIPS:  # QL_MIPS32
            if self.archendian == QL_ENDIAN.EB:
                md = Cs(CS_ARCH_MIPS, CS_MODE_MIPS32 + CS_MODE_BIG_ENDIAN)
            else:
                md = Cs(CS_ARCH_MIPS, CS_MODE_MIPS32 + CS_MODE_LITTLE_ENDIAN)

        else:
            raise QlErrorArch("[!] Unknown arch defined in utils.py (debug output mode)")

        return md
    
    def create_assembler(self):
        if self.archtype == QL_ARCH.ARM:  # QL_ARM
            reg_cpsr = self.reg.cpsr
            mode = KS_MODE_ARM
            if self.archendian == QL_ENDIAN.EB:
                reg_cpsr_v = 0b100000
                # reg_cpsr_v = 0b000000
            else:
                reg_cpsr_v = 0b100000

            if reg_cpsr & reg_cpsr_v != 0:
                mode = KS_MODE_THUMB

            if self.archendian == QL_ENDIAN.EB:
                ks = Ks(KS_ARCH_ARM, mode)
                # md = Cs(CS_ARCH_ARM, mode + CS_MODE_BIG_ENDIAN)
            else:
                ks = Ks(KS_ARCH_ARM, mode)

        elif self.archtype == QL_ARCH.ARM_THUMB:
            ks = Ks(KS_ARCH_ARM, KS_MODE_THUMB)

        elif self.archtype == QL_ARCH.X86:  # QL_X86
            ks = Ks(KS_ARCH_X86, KS_MODE_32)

        elif self.archtype == QL_ARCH.X8664:  # QL_X86_64
            ks = Ks(KS_ARCH_X86, KS_MODE_64)

        elif self.archtype == QL_ARCH.ARM64:  # QL_ARM64
            ks = Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN)

        elif self.archtype == QL_ARCH.A8086:  # QL_A8086
            ks = Ks(KS_ARCH_X86, KS_MODE_16)

        elif self.archtype == QL_ARCH.MIPS:  # QL_MIPS32
            if self.archendian == QL_ENDIAN.EB:
                ks = Ks(KS_ARCH_MIPS, KS_MODE_MIPS32 + KS_MODE_BIG_ENDIAN)
            else:
                ks = Ks(KS_ARCH_MIPS, KS_MODE_MIPS32 + KS_MODE_LITTLE_ENDIAN)

        else:
            raise QlErrorArch("[!] Unknown arch defined in utils.py (debug output mode)")

        return ks

class QlFileDes:
    def __init__(self, init):
        self.__fds = init

    def __getitem__(self, idx):
        return self.__fds[idx]

    def __setitem__(self, idx, val):
        self.__fds[idx] = val

    def __iter__(self):
        return iter(self.__fds)

    def __repr__(self):
        return repr(self.__fds)
    
    def save(self):
        return self.__fds

    def restore(self, fds):
        self.__fds = fds




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

pointer_size = 8

EFI_MAX_BIT = 0x8000000000000000
EFI_SUCCESS = 0
EFI_LOAD_ERROR = EFI_MAX_BIT | 1
EFI_INVALID_PARAMETER = EFI_MAX_BIT | 2
EFI_UNSUPPORTED = EFI_MAX_BIT | 3
EFI_BAD_BUFFER_SIZE = EFI_MAX_BIT | 4
EFI_BUFFER_TOO_SMALL = EFI_MAX_BIT | 5
EFI_NOT_READY = EFI_MAX_BIT | 6
EFI_DEVICE_ERROR = EFI_MAX_BIT | 7
EFI_WRITE_PROTECTED = EFI_MAX_BIT | 8
EFI_OUT_OF_RESOURCES = EFI_MAX_BIT | 9
EFI_VOLUME_CORRUPTED = EFI_MAX_BIT | 10
EFI_VOLUME_FULL = EFI_MAX_BIT | 11
EFI_NO_MEDIA = EFI_MAX_BIT | 12
EFI_MEDIA_CHANGED = EFI_MAX_BIT | 13
EFI_NOT_FOUND = EFI_MAX_BIT | 14
EFI_ACCESS_DENIED = EFI_MAX_BIT | 15
EFI_NO_RESPONSE = EFI_MAX_BIT | 16
EFI_NO_MAPPING = EFI_MAX_BIT | 17
EFI_TIMEOUT = EFI_MAX_BIT | 18
EFI_NOT_STARTED = EFI_MAX_BIT | 19
EFI_ALREADY_STARTED = EFI_MAX_BIT | 20
EFI_ABORTED = EFI_MAX_BIT | 21
EFI_ICMP_ERROR = EFI_MAX_BIT | 22
EFI_TFTP_ERROR = EFI_MAX_BIT | 23
EFI_PROTOCOL_ERROR = EFI_MAX_BIT | 24
EFI_INCOMPATIBLE_VERSION = EFI_MAX_BIT | 25
EFI_SECURITY_VIOLATION = EFI_MAX_BIT | 26
EFI_CRC_ERROR = EFI_MAX_BIT | 27
EFI_END_OF_MEDIA = EFI_MAX_BIT | 28
EFI_END_OF_FILE = EFI_MAX_BIT | 31
EFI_INVALID_LANGUAGE = EFI_MAX_BIT | 32
EFI_WARN_UNKNOWN_GLYPH = EFI_MAX_BIT | 1
EFI_WARN_DELETE_FAILURE = EFI_MAX_BIT | 2
EFI_WARN_WRITE_FAILURE = EFI_MAX_BIT | 3
EFI_WARN_BUFFER_TOO_SMALL = EFI_MAX_BIT | 4

SEARCHTYPE_AllHandles = 0
SEARCHTYPE_ByRegisterNotify = 1
SEARCHTYPE_ByProtocol = 2




#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)

import struct
import time
from qiling.os.windows.const import *
from qiling.os.const import *
from qiling.os.windows.fncc import *
from qiling.os.windows.utils import *
from qiling.os.windows.thread import *
from qiling.os.windows.handle import *
from qiling.exception import *


dllname = 'kernel32_dll'

# HWND WINAPI GetConsoleWindow(void);
@winsdkapi(cc=STDCALL, dllname=dllname)
def hook_GetConsoleWindow(ql, address, params):
    handle = Handle(name="console_window")
    ql.os.handle_manager.append(handle)
    return handle.id



#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

from .utils import *

def hook_EndOfExecution(ql):
    if ql.os.notify_after_module_execution(ql, len(ql.loader.modules)):
        return
    ql.loader.restore_runtime_services()
    if check_and_notify_protocols(ql):
        return
    if len(ql.loader.modules) < 1:
        if ql.loader.unload_modules():
            return
        ql.nprint(f'[+] No more modules to run')
        ql.emu_stop()
    else:
        ql.loader.execute_next_module()




#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org) 

from uuid import UUID
from binascii import crc32
from qiling.const import *

from .utils import *
from .type64 import *
from .mm_system_table_type import *
from .shutdown import *
from .fncc import *
from .const import *
from qiling.os.const import *

@dxeapi(params={
    "NewTpl": ULONGLONG,
})
def hook_RaiseTPL(ql, address, params):
    tpl = ql.loader.tpl
    ql.loader.tpl = params["NewTpl"]
    return tpl

@dxeapi(params={
    "OldTpl": ULONGLONG,
})
def hook_RestoreTPL(ql, address, params):
    ql.loader.tpl = params["OldTpl"]

@dxeapi(params={
    "type": ULONGLONG,
    "MemoryType": ULONGLONG,
    "Pages": ULONGLONG,
    "Memory": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_AllocatePages(ql, address, params):
    AllocateAnyPages = 0
    AllocateMaxAddress = 1
    AllocateAddress = 2
    PageSize = 4096
    if params['type'] == AllocateAddress:
        address =  read_int64(ql, params["Memory"])
        ql.mem.map(address, params["Pages"]*PageSize)
    else:
        address = ql.os.heap.alloc(params["Pages"]*PageSize)
        write_int64(ql, params["Memory"], address)
    return EFI_SUCCESS

@dxeapi(params={
    "Memory": ULONGLONG,
    "Pages": ULONGLONG,
})
def hook_FreePages(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(ctypes.c_uint64)
    "a1": POINTER, #POINTER_T(struct_EFI_MEMORY_DESCRIPTOR)
    "a2": POINTER, #POINTER_T(ctypes.c_uint64)
    "a3": POINTER, #POINTER_T(ctypes.c_uint64)
    "a4": POINTER, #POINTER_T(ctypes.c_uint32)
})
def hook_GetMemoryMap(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "PoolType": UINT,
    "Size": UINT,
    "Buffer": POINTER,
})
def hook_AllocatePool(ql, address, params):
    address = ql.os.heap.alloc(params["Size"])
    write_int64(ql, params["Buffer"], address)
    return EFI_SUCCESS if address else EFI_OUT_OF_RESOURCES

@dxeapi(params={
    "Buffer": POINTER, #POINTER_T(None)
})
def hook_FreePool(ql, address, params):
    address = params["Buffer"]
    freed = ql.os.heap.free(address)
    return EFI_SUCCESS if freed else EFI_INVALID_PARAMETER 

@dxeapi(params={
    "Type": UINT,
    "NotifyTpl": UINT,
    "NotifyFunction": POINTER,
    "NotifyContext": POINTER,
    "Event": POINTER})
def hook_CreateEvent(ql, address, params):
    return CreateEvent(ql, address, params)

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": ULONGLONG,
    "a2": ULONGLONG,
})
def hook_SetTimer(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(POINTER_T(None))
    "a2": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_WaitForEvent(ql, address, params):
    return EFI_SUCCESS

def SignalEvent(ql, event_id):
    if event_id in ql.loader.events:
        event = ql.loader.events[event_id]
        if not event["Set"]:
            event["Set"] = True
            notify_func = event["NotifyFunction"]
            notify_context = event["NotifyContext"]
            ql.loader.notify_list.append((event_id, notify_func, notify_context))
        return EFI_SUCCESS
    else:
        return EFI_INVALID_PARAMETER

@dxeapi(params={
    "Event": POINTER, #POINTER_T(None)
})
def hook_SignalEvent(ql, address, params):
    event_id = params["Event"]
    return SignalEvent(ql, event_id)

@dxeapi(params={
    "Event": POINTER, #POINTER_T(None)
})
def hook_CloseEvent(ql, address, params):
    del ql.loader.events[params["Event"]]
    return EFI_SUCCESS

@dxeapi(params={
    "Event": POINTER, #POINTER_T(None)
})
def hook_CheckEvent(ql, address, params):
    return EFI_SUCCESS if ql.loader.events[params["Event"]]["Set"] else EFI_NOT_READY

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(POINTER_T(None))
    "Protocol": GUID,
    "InterfaceType": ULONGLONG,
    "Interface": POINTER, #POINTER_T(None)
})
def hook_InstallProtocolInterface(ql, address, params):
    dic = {}
    handle = read_int64(ql, params["Handle"])
    if handle == 0:
        handle = ql.os.heap.alloc(1)
    if handle in ql.loader.handle_dict:
        dic = ql.loader.handle_dict[handle]
    dic[params["Protocol"]] = params["Interface"]
    ql.loader.handle_dict[handle] = dic
    check_and_notify_protocols(ql)
    write_int64(ql, params["Handle"], handle)
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "OldInterface": POINTER, #POINTER_T(None)
    "NewInterface": POINTER, #POINTER_T(None)
})
def hook_ReinstallProtocolInterface(ql, address, params):
    handle = params["Handle"]
    if handle not in ql.loader.handle_dict:
        return EFI_NOT_FOUND
    dic = ql.loader.handle_dict[handle]
    protocol = params["Protocol"]
    if protocol not in dic:
        return EFI_NOT_FOUND
    dic[protocol] = params["NewInterface"]
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "Interface": POINTER, #POINTER_T(None)
})
def hook_UninstallProtocolInterface(ql, address, params):
    handle = params["Handle"]
    if handle not in ql.loader.handle_dict:
        return EFI_NOT_FOUND
    dic = ql.loader.handle_dict[handle]
    protocol = params["Protocol"]
    if protocol not in dic:
        return EFI_NOT_FOUND
    del dic[protocol]
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "Interface": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_HandleProtocol(ql, address, params):
    handle = params["Handle"]
    protocol = params["Protocol"]
    interface = params['Interface']
    if handle in ql.loader.handle_dict:
        if protocol in ql.loader.handle_dict[handle]:
            write_int64(ql, interface, ql.loader.handle_dict[handle][protocol])
            return EFI_SUCCESS
    return EFI_NOT_FOUND

@dxeapi(params={
    "Protocol": GUID,
    "Event": POINTER,
    "Registration": POINTER})
def hook_RegisterProtocolNotify(ql, address, params):
    if params['Event'] in ql.loader.events:
        ql.loader.events[params['Event']]['Guid'] = params["Protocol"]
        check_and_notify_protocols(ql)
        return EFI_SUCCESS
    return EFI_INVALID_PARAMETER



@dxeapi(params={
    "SearchType": ULONGLONG,
    "Protocol": GUID,
    "SearchKey": POINTER, #POINTER_T(None)
    "BufferSize": POINTER, #POINTER_T(ctypes.c_uint64)
    "Buffer": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_LocateHandle(ql, address, params):
    buffer_size, handles = LocateHandles(ql, address, params)
    if len(handles) == 0:
        return EFI_NOT_FOUND
    ret = EFI_BUFFER_TOO_SMALL
    if read_int64(ql, params["BufferSize"]) >= buffer_size:
        ptr = params["Buffer"]
        for handle in handles:
            write_int64(ql, ptr, handle)
            ptr += pointer_size
        ret = EFI_SUCCESS
    write_int64(ql, params["BufferSize"], buffer_size)
    return ret
    

@dxeapi(params={
    "a0": GUID,
    "a1": POINTER, #POINTER_T(POINTER_T(struct_EFI_DEVICE_PATH_PROTOCOL))
    "a2": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_LocateDevicePath(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "Guid": GUID,
    "Table": POINTER, #POINTER_T(None)
})
def hook_InstallConfigurationTable(ql, address, params):
    guid = params["Guid"]
    uuid = UUID(guid)
    efi_configuration_table = EFI_CONFIGURATION_TABLE()
    if guid in ql.loader.efi_configuration_table:
        index = ql.loader.efi_configuration_table.index(guid)
    else:
        index = len(ql.loader.efi_configuration_table)
        ql.loader.efi_configuration_table.append(guid)
    
    address = (index * ctypes.sizeof(efi_configuration_table)) + ql.loader.efi_configuration_table_ptr
    efi_configuration_table.VendorGuid.Data1 = uuid.fields[0]
    efi_configuration_table.VendorGuid.Data2 = uuid.fields[1]
    efi_configuration_table.VendorGuid.Data3 = uuid.fields[2]
    guid_last_bytes = uuid.bytes_le[8:]
    for i in range(len(guid_last_bytes)):
        efi_configuration_table.VendorGuid.Data4[i] = guid_last_bytes[i]
    
    efi_configuration_table.VendorTable = params["Table"]
    ql.mem.write(address, convert_struct_to_bytes(efi_configuration_table))
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": POINTER, #POINTER_T(None)
    "a2": POINTER, #POINTER_T(struct_EFI_DEVICE_PATH_PROTOCOL)
    "a3": POINTER, #POINTER_T(None)
    "a4": ULONGLONG,
    "a5": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_LoadImage(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": POINTER, #POINTER_T(ctypes.c_uint64)
    "a2": POINTER, #POINTER_T(POINTER_T(ctypes.c_uint16))
})
def hook_StartImage(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": POINTER, #POINTER_T(ctypes.c_uint16)
})
def hook_Exit(ql, address, params):
    ql.nprint(f'hook_Exit')
    ql.uc.emu_stop()
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
})
def hook_UnloadImage(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": ULONGLONG,
})
def hook_ExitBootServices(ql, address, params):
    ql.nprint(f'hook_ExitBootServices')
    ql.uc.emu_stop()
    return EFI_SUCCESS

@dxeapi(params={
    "Count": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_GetNextMonotonicCount(ql, address, params):
    ql.os.monotonic_count += 1
    write_int64(ql, params["Count"], ql.os.monotonic_count)
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
})
def hook_Stall(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": ULONGLONG,
    "a1": ULONGLONG,
    "a2": ULONGLONG,
    "a3": POINTER, #POINTER_T(ctypes.c_uint16)
})
def hook_SetWatchdogTimer(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": POINTER, #POINTER_T(POINTER_T(None))
    "a2": POINTER, #POINTER_T(struct_EFI_DEVICE_PATH_PROTOCOL)
    "a3": ULONGLONG,
})
def hook_ConnectController(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": POINTER, #POINTER_T(None)
    "a2": POINTER, #POINTER_T(None)
})
def hook_DisconnectController(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "Interface": POINTER, #POINTER_T(POINTER_T(None))
    "AgentHandle": POINTER, #POINTER_T(None)
    "ControllerHandle": POINTER, #POINTER_T(None)
    "Attributes": UINT,
})
def hook_OpenProtocol(ql, address, params):
    return LocateProtocol(ql, address, params)

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "AgentHandle": POINTER, #POINTER_T(None)
    "ControllerHandle": POINTER, #POINTER_T(None)
})
def hook_CloseProtocol(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
    "Protocol": GUID,
    "EntryBuffer": POINTER, #POINTER_T(POINTER_T(struct_EFI_OPEN_PROTOCOL_INFORMATION_ENTRY))
    "EntryCount": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_OpenProtocolInformation(ql, address, params):
    return EFI_NOT_FOUND

@dxeapi(params={
    "a0": POINTER, #POINTER_T(None)
    "a1": GUID,
    "a2": POINTER, #POINTER_T(ctypes.c_uint64)
})
def hook_ProtocolsPerHandle(ql, address, params):
    return EFI_SUCCESS

@dxeapi(params={
    "SearchType": ULONGLONG,
    "Protocol": GUID,
    "SearchKey": POINTER, #POINTER_T(None)
    "NoHandles": POINTER, #POINTER_T(ctypes.c_uint64)
    "Buffer": POINTER, #POINTER_T(POINTER_T(POINTER_T(None)))
})
def hook_LocateHandleBuffer(ql, address, params):
    buffer_size, handles = LocateHandles(ql, address, params)
    write_int64(ql, params["NoHandles"], len(handles))
    if len(handles) == 0:
        return EFI_NOT_FOUND
    address = ql.os.heap.alloc(buffer_size)
    write_int64(ql, params["Buffer"], address)
    if address == 0:
        return EFI_OUT_OF_RESOURCES
    for handle in handles:
        write_int64(ql, address, handle)
        address += pointer_size
    return EFI_SUCCESS

@dxeapi(params={
    "Protocol": GUID,
    "Registration": POINTER, #POINTER_T(None)
    "Interface": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_LocateProtocol(ql, address, params):
    return LocateProtocol(ql, address, params)

@dxeapi(params={
    "Handle": POINTER})
def hook_InstallMultipleProtocolInterfaces(ql, address, params):
    handle = read_int64(ql, params["Handle"])
    if handle == 0:
        handle = ql.os.heap.alloc(1)
    ql.nprint(f'hook_InstallMultipleProtocolInterfaces {handle:x}')
    dic = {}
    if handle in ql.loader.handle_dict:
        dic = ql.loader.handle_dict[handle]
    
    index = 1
    while ql.os.get_param_by_index(index) != 0:
        GUID_ptr = ql.os.get_param_by_index(index)
        protocol_ptr = ql.os.get_param_by_index(index+1)
        GUID = str(ql.os.read_guid(GUID_ptr))
        ql.nprint(f'\t {GUID}, {protocol_ptr:x}')
        dic[GUID] = protocol_ptr
        index +=2
    ql.loader.handle_dict[handle] = dic
    check_and_notify_protocols(ql)
    write_int64(ql, params["Handle"], handle)
    return EFI_SUCCESS

@dxeapi(params={
    "Handle": POINTER, #POINTER_T(None)
})
def hook_UninstallMultipleProtocolInterfaces(ql, address, params):
    handle = params["Handle"]
    ql.nprint(f'hook_UninstallMultipleProtocolInterfaces {handle:x}')
    if handle not in ql.loader.handle_dict:
        return EFI_NOT_FOUND
    index = 1
    while ql.os.get_param_by_index(index) != 0:
        GUID_ptr = ql.os.get_param_by_index(index)
        protocol_ptr = ql.os.get_param_by_index(index+1)
        GUID = str(ql.os.read_guid(GUID_ptr))
        ql.nprint(f'\t {GUID}, {protocol_ptr:x}')
        dic = ql.loader.handle_dict[handle]
        if GUID not in dic:
            return EFI_INVALID_PARAMETER
        del dic[GUID]
        index +=2
    return EFI_SUCCESS

@dxeapi(params={
    "Data": POINTER, #POINTER_T(None)
    "DataSize": ULONGLONG,
    "Crc32": POINTER, #POINTER_T(ctypes.c_uint32)
})
def hook_CalculateCrc32(ql, address, params):
    data = bytes(ql.mem.read(params['Data'], params['DataSize']))
    write_int32(ql, params['Crc32'], crc32(data))
    return EFI_SUCCESS

@dxeapi(params={
    "Destination": POINTER, #POINTER_T(None)
    "Source": POINTER, #POINTER_T(None)
    "Length": ULONGLONG,
})
def hook_CopyMem(ql, address, params):
    data = bytes(ql.mem.read(params['Source'], params['Length']))
    ql.mem.write(params['Destination'], data)

@dxeapi(params={
    "Buffer": POINTER, #POINTER_T(None)
    "Size": ULONGLONG,
    "Value": BYTE,
})
def hook_SetMem(ql, address, params):
    ptr = params["Buffer"]
    value = struct.pack('B',params["Value"] & 0xff)
    for i in range(0, params["Size"]):
        ql.mem.write(ptr+i, value)

@dxeapi(params={
    "Type": UINT,
    "NotifyTpl": ULONGLONG,
    "NotifyFunction": POINTER, #POINTER_T(ctypes.CFUNCTYPE(None, POINTER_T(None), POINTER_T(None)))
    "NotifyContext": POINTER, #POINTER_T(None)
    "EventGroup": GUID,
    "Event": POINTER, #POINTER_T(POINTER_T(None))
})
def hook_CreateEventEx(ql, address, params):
    return CreateEvent(ql, address, params)


def CreateEvent(ql, address, params):
    event_id = len(ql.loader.events)
    event_dic = {"NotifyFunction": params["NotifyFunction"], "NotifyContext": params["NotifyContext"], "Guid": "", "Set": False}
    if "EventGroup" in params:
        event_dic["EventGroup"] =  params["EventGroup"]
    
    ql.loader.events[event_id] = event_dic
    write_int64(ql, params["Event"], event_id)
    return EFI_SUCCESS

def create_EFI_MM_SYSTEM_TABLE(ql, start_ptr):
    efi_mm_system_table = EFI_MM_SYSTEM_TABLE()

    ptr = start_ptr
    efi_mm_system_table.MmAllocatePages = ptr
    ql.hook_address(hook_AllocatePages, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmFreePages = ptr
    ql.hook_address(hook_FreePages, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmAllocatePool = ptr
    ql.hook_address(hook_AllocatePool, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmFreePool = ptr
    ql.hook_address(hook_FreePool, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmInstallProtocolInterface = ptr
    ql.hook_address(hook_InstallProtocolInterface, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmUninstallProtocolInterface = ptr
    ql.hook_address(hook_UninstallProtocolInterface, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmHandleProtocol = ptr
    ql.hook_address(hook_HandleProtocol, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmRegisterProtocolNotify = ptr
    ql.hook_address(hook_RegisterProtocolNotify, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmLocateHandle = ptr
    ql.hook_address(hook_LocateHandle, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmInstallConfigurationTable = ptr
    ql.hook_address(hook_InstallConfigurationTable, ptr)
    ptr += pointer_size
    efi_mm_system_table.MmLocateProtocol = ptr
    ql.hook_address(hook_LocateProtocol, ptr)
    ptr += pointer_size
    return (ptr, efi_mm_system_table)


def hook_EFI_BOOT_SERVICES(ql, start_ptr):
    ql.os.monotonic_count = 0

    efi_boot_services = EFI_BOOT_SERVICES()

    ptr = start_ptr
    efi_boot_services.RaiseTPL = ptr
    ql.hook_address(hook_RaiseTPL, ptr)
    ptr += pointer_size
    efi_boot_services.RestoreTPL = ptr
    ql.hook_address(hook_RestoreTPL, ptr)
    ptr += pointer_size
    efi_boot_services.AllocatePages = ptr
    ql.hook_address(hook_AllocatePages, ptr)
    ptr += pointer_size
    efi_boot_services.FreePages = ptr
    ql.hook_address(hook_FreePages, ptr)
    ptr += pointer_size
    efi_boot_services.GetMemoryMap = ptr
    ql.hook_address(hook_GetMemoryMap, ptr)
    ptr += pointer_size
    efi_boot_services.AllocatePool = ptr
    ql.hook_address(hook_AllocatePool, ptr)
    ptr += pointer_size
    efi_boot_services.FreePool = ptr
    ql.hook_address(hook_FreePool, ptr)
    ptr += pointer_size
    efi_boot_services.CreateEvent = ptr
    ql.hook_address(hook_CreateEvent, ptr)
    ptr += pointer_size
    efi_boot_services.SetTimer = ptr
    ql.hook_address(hook_SetTimer, ptr)
    ptr += pointer_size
    efi_boot_services.WaitForEvent = ptr
    ql.hook_address(hook_WaitForEvent, ptr)
    ptr += pointer_size
    efi_boot_services.SignalEvent = ptr
    ql.hook_address(hook_SignalEvent, ptr)
    ptr += pointer_size
    efi_boot_services.CloseEvent = ptr
    ql.hook_address(hook_CloseEvent, ptr)
    ptr += pointer_size
    efi_boot_services.CheckEvent = ptr
    ql.hook_address(hook_CheckEvent, ptr)
    ptr += pointer_size
    efi_boot_services.InstallProtocolInterface = ptr
    ql.hook_address(hook_InstallProtocolInterface, ptr)
    ptr += pointer_size
    efi_boot_services.ReinstallProtocolInterface = ptr
    ql.hook_address(hook_ReinstallProtocolInterface, ptr)
    ptr += pointer_size
    efi_boot_services.UninstallProtocolInterface = ptr
    ql.hook_address(hook_UninstallProtocolInterface, ptr)
    ptr += pointer_size
    efi_boot_services.HandleProtocol = ptr
    ql.hook_address(hook_HandleProtocol, ptr)
    ptr += pointer_size
    efi_boot_services.RegisterProtocolNotify = ptr
    ql.hook_address(hook_RegisterProtocolNotify, ptr)
    ptr += pointer_size
    efi_boot_services.LocateHandle = ptr
    ql.hook_address(hook_LocateHandle, ptr)
    ptr += pointer_size
    efi_boot_services.LocateDevicePath = ptr
    ql.hook_address(hook_LocateDevicePath, ptr)
    ptr += pointer_size
    efi_boot_services.InstallConfigurationTable = ptr
    ql.hook_address(hook_InstallConfigurationTable, ptr)
    ptr += pointer_size
    efi_boot_services.LoadImage = ptr
    ql.hook_address(hook_LoadImage, ptr)
    ptr += pointer_size
    efi_boot_services.StartImage = ptr
    ql.hook_address(hook_StartImage, ptr)
    ptr += pointer_size
    efi_boot_services.Exit = ptr
    ql.hook_address(hook_Exit, ptr)
    ptr += pointer_size
    efi_boot_services.UnloadImage = ptr
    ql.hook_address(hook_UnloadImage, ptr)
    ptr += pointer_size
    efi_boot_services.ExitBootServices = ptr
    ql.hook_address(hook_ExitBootServices, ptr)
    ptr += pointer_size
    efi_boot_services.GetNextMonotonicCount = ptr
    ql.hook_address(hook_GetNextMonotonicCount, ptr)
    ptr += pointer_size
    efi_boot_services.Stall = ptr
    ql.hook_address(hook_Stall, ptr)
    ptr += pointer_size
    efi_boot_services.SetWatchdogTimer = ptr
    ql.hook_address(hook_SetWatchdogTimer, ptr)
    ptr += pointer_size
    efi_boot_services.ConnectController = ptr
    ql.hook_address(hook_ConnectController, ptr)
    ptr += pointer_size
    efi_boot_services.DisconnectController = ptr
    ql.hook_address(hook_DisconnectController, ptr)
    ptr += pointer_size
    efi_boot_services.OpenProtocol = ptr
    ql.hook_address(hook_OpenProtocol, ptr)
    ptr += pointer_size
    efi_boot_services.CloseProtocol = ptr
    ql.hook_address(hook_CloseProtocol, ptr)
    ptr += pointer_size
    efi_boot_services.OpenProtocolInformation = ptr
    ql.hook_address(hook_OpenProtocolInformation, ptr)
    ptr += pointer_size
    efi_boot_services.ProtocolsPerHandle = ptr
    ql.hook_address(hook_ProtocolsPerHandle, ptr)
    ptr += pointer_size
    efi_boot_services.LocateHandleBuffer = ptr
    ql.hook_address(hook_LocateHandleBuffer, ptr)
    ptr += pointer_size
    efi_boot_services.LocateProtocol = ptr
    ql.hook_address(hook_LocateProtocol, ptr)
    ptr += pointer_size
    efi_boot_services.InstallMultipleProtocolInterfaces = ptr
    ql.hook_address(hook_InstallMultipleProtocolInterfaces, ptr)
    ptr += pointer_size
    efi_boot_services.UninstallMultipleProtocolInterfaces = ptr
    ql.hook_address(hook_UninstallMultipleProtocolInterfaces, ptr)
    ptr += pointer_size
    efi_boot_services.CalculateCrc32 = ptr
    ql.hook_address(hook_CalculateCrc32, ptr)
    ptr += pointer_size
    efi_boot_services.CopyMem = ptr
    ql.hook_address(hook_CopyMem, ptr)
    ptr += pointer_size
    efi_boot_services.SetMem = ptr
    ql.hook_address(hook_SetMem, ptr)
    ptr += pointer_size
    efi_boot_services.CreateEventEx = ptr
    ql.hook_address(hook_CreateEventEx, ptr)
    ptr += pointer_size
    return (ptr, efi_boot_services)




#!/usr/bin/env python3
#
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
# Built on top of Unicorn emulator (www.unicorn-engine.org)


from qiling.const import *
from qiling.os.linux.thread import *
from qiling.const import *
from qiling.os.posix.filestruct import *
from qiling.os.filestruct import *
from qiling.os.posix.const_mapping import *
from qiling.exception import *

def ql_syscall_ptrace(ql, request, pid, addr, data, *args, **kw):
    regreturn = 0
    ql.nprint("ptrace(0x%x, 0x%x, 0x%x, 0x%x) = %d" % (request, pid, addr, data, regreturn))
    ql.os.definesyscall_return(regreturn)
